# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import sqlite3
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_unix_to_iso

from ..base import IOSExtraction

TCC_BACKUP_IDS = [
    "64d0019cb3d46bfc8cce545a8ba54b93e7ea9347",
]
TCC_ROOT_PATHS = [
    "private/var/mobile/Library/TCC/TCC.db",
]

AUTH_VALUE_OLD = {0: "denied", 1: "allowed"}
AUTH_VALUES = {
    0: "denied",
    1: "unknown",
    2: "allowed",
    3: "limited",
}
AUTH_REASONS = {
    1: "error",
    2: "user_consent",
    3: "user_set",
    4: "system_set",
    5: "service_policy",
    6: "mdm_policy",
    7: "override_policy",
    8: "missing_usage_string",
    9: "prompt_timeout",
    10: "preflight_unknown",
    11: "entitled",
    12: "app_type_policy",
}


class TCC(IOSExtraction):
    """This module extracts records from the TCC.db SQLite database."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        if "last_modified" in record:
            if "allowed_value" in record:
                msg = (
                    f"Access to {record['service']} by {record['client']} "
                    f"{record['allowed_value']}"
                )
            else:
                msg = (
                    f"Access to {record['service']} by {record['client']} "
                    f"{record['auth_value']}"
                )

            return {
                "timestamp": record["last_modified"],
                "module": self.__class__.__name__,
                "event": "AccessRequest",
                "data": msg,
            }

        return {}

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_process(result["client"])
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

    def process_db(self, file_path):
        conn = self._open_sqlite_db(file_path)
        cur = conn.cursor()
        db_version = "v3"
        try:
            cur.execute(
                """SELECT
                service, client, client_type, auth_value,
                auth_reason, last_modified
            FROM access;"""
            )
        except sqlite3.OperationalError:
            # v2 version
            try:
                cur.execute(
                    """SELECT
                    service, client, client_type, allowed,
                    prompt_count, last_modified
                    FROM access;"""
                )
                db_version = "v2"
            except sqlite3.OperationalError:
                try:
                    cur.execute(
                        """SELECT
                        service, client, client_type, allowed,
                        prompt_count
                        FROM access;"""
                    )
                    db_version = "v1"
                except sqlite3.OperationalError as e:
                    self.log.error(f"Error parsing TCC database: {e}")

        for row in cur:
            service = row[0]
            client = row[1]
            client_type = row[2]
            client_type_desc = "bundle_id" if client_type == 0 else "absolute_path"
            if db_version == "v3":
                auth_value = row[3]
                auth_value_desc = AUTH_VALUES.get(auth_value, "")
                auth_reason = row[4]
                auth_reason_desc = AUTH_REASONS.get(auth_reason, "unknown")
                last_modified = convert_unix_to_iso(row[5])

                if service in ["kTCCServiceMicrophone", "kTCCServiceCamera"]:
                    device = (
                        "microphone" if service == "kTCCServiceMicrophone" else "camera"
                    )
                    self.log.info(
                        'Found client "%s" with access %s to %s on %s by %s',
                        client,
                        auth_value_desc,
                        device,
                        last_modified,
                        auth_reason_desc,
                    )

                self.results.append(
                    {
                        "service": service,
                        "client": client,
                        "client_type": client_type_desc,
                        "auth_value": auth_value_desc,
                        "auth_reason_desc": auth_reason_desc,
                        "last_modified": last_modified,
                    }
                )
            else:
                allowed_value = row[3]
                allowed_desc = AUTH_VALUE_OLD.get(allowed_value, "")
                prompt_count = row[4]

                if db_version == "v2":
                    last_modified = convert_unix_to_iso(row[5])
                    if service in ["kTCCServiceMicrophone", "kTCCServiceCamera"]:
                        device = "camera"
                        if service == "kTCCServiceMicrophone":
                            device = "microphone"

                        self.log.info(
                            'Found client "%s" with access %s to %s at %s',
                            client,
                            allowed_desc,
                            device,
                            last_modified,
                        )

                    self.results.append(
                        {
                            "service": service,
                            "client": client,
                            "client_type": client_type_desc,
                            "allowed_value": allowed_desc,
                            "prompt_count": prompt_count,
                            "last_modified": last_modified,
                        }
                    )
                else:
                    if service in ["kTCCServiceMicrophone", "kTCCServiceCamera"]:
                        device = "camera"
                        if service == "kTCCServiceMicrophone":
                            device = "microphone"

                        self.log.info(
                            'Found client "%s" with access %s to %s',
                            client,
                            allowed_desc,
                            device,
                        )

                    self.results.append(
                        {
                            "service": service,
                            "client": client,
                            "client_type": client_type_desc,
                            "allowed_value": allowed_desc,
                            "prompt_count": prompt_count,
                        }
                    )

        cur.close()
        conn.close()

    def run(self) -> None:
        self._find_ios_database(backup_ids=TCC_BACKUP_IDS, root_paths=TCC_ROOT_PATHS)
        self.log.info("Found TCC database at path: %s", self.file_path)

        self.process_db(self.file_path)

        self.log.info("Extracted a total of %d TCC items", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2026 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

"""Stable functions and modules which plugins can import.

Anything else in mvt can be imported too, and may change
between releases without notice.
"""

from mvt.android.modules.androidqf.base import AndroidQFModule
from mvt.android.modules.backup.base import BackupModule as AndroidBackupModule
from mvt.android.modules.bugreport.base import BugReportModule
from mvt.common.alerts import Alert, AlertLevel
from mvt.common.command import Command
from mvt.common.config import settings
from mvt.common.module import DatabaseCorruptedError, DatabaseNotFoundError, MVTModule
from mvt.common.module_loader import get_plugin_logger
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.plugin_config import (
    MVTPluginSettings,
    PluginConfigLoadError,
    plugin_config_path,
    plugin_data_folder,
    plugin_env_prefix,
)
from mvt.common.utils import (
    convert_chrometime_to_datetime,
    convert_datetime_to_iso,
    convert_mactime_to_datetime,
    convert_mactime_to_iso,
    convert_unix_to_iso,
    convert_unix_to_utc_datetime,
)
from mvt.common.version import MVT_VERSION
from mvt.ios.modules.base import IOSExtraction
from mvt.ios.modules.sysdiagnose.base import SysdiagnoseExtraction

__all__ = [
    # Classes a plugin subclasses.
    "MVTModule",
    "IOSExtraction",
    "SysdiagnoseExtraction",
    "AndroidQFModule",
    "AndroidBackupModule",
    "BugReportModule",
    "Command",
    # Results and alerts.
    "ModuleAtomicResult",
    "ModuleResults",
    "ModuleSerializedResult",
    "Alert",
    "AlertLevel",
    # Errors a module raises.
    "DatabaseNotFoundError",
    "DatabaseCorruptedError",
    # Settings.
    "settings",
    "MVTPluginSettings",
    "PluginConfigLoadError",
    "plugin_config_path",
    "plugin_data_folder",
    "plugin_env_prefix",
    # Logging.
    "get_plugin_logger",
    # Timestamps.
    "convert_chrometime_to_datetime",
    "convert_datetime_to_iso",
    "convert_mactime_to_datetime",
    "convert_mactime_to_iso",
    "convert_unix_to_iso",
    "convert_unix_to_utc_datetime",
    # MVT's version.
    "MVT_VERSION",
]



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2026 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
# https://license.mvt.re/1.1/

"""Password prompts with masked keyboard feedback.

This is based on the ``echo_char`` support added to :mod:`getpass` in Python
3.14. MVT supports Python 3.10 and later, so it cannot use that API directly.
"""

import contextlib
import io
import os
import sys
import warnings
from typing import TextIO


def prompt_password(prompt: str) -> str:
    """Return a password while displaying an asterisk for each character."""
    try:
        import termios

        return _unix_getpass(prompt, termios)
    except ImportError:
        try:
            import msvcrt

            return _windows_getpass(prompt, msvcrt)
        except ImportError:
            return _fallback_getpass(prompt)


def _unix_getpass(prompt: str, termios) -> str:
    password = None
    input_stream: TextIO
    output_stream: TextIO
    with contextlib.ExitStack() as stack:
        try:
            fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
            tty = io.FileIO(fd, "w+")
            stack.enter_context(tty)
            input_stream = io.TextIOWrapper(tty)
            stack.enter_context(input_stream)
            output_stream = input_stream
        except OSError:
            stack.close()
            try:
                fd = sys.stdin.fileno()
            except (AttributeError, ValueError):
                return _fallback_getpass(prompt)
            input_stream = sys.stdin
            output_stream = sys.stderr

        try:
            old = termios.tcgetattr(fd)
            new = old[:]
            new[3] &= ~termios.ECHO
            new[3] &= ~termios.ICANON
            try:
                termios.tcsetattr(fd, termios.TCSAFLUSH, new)
                password = _readline_with_asterisks(output_stream, input_stream, prompt)
            finally:
                termios.tcsetattr(fd, termios.TCSAFLUSH, old)
                output_stream.flush()
        except termios.error:
            if password is not None:
                raise
            password = _fallback_getpass(prompt)

        output_stream.write("\n")
        return password


def _windows_getpass(prompt: str, msvcrt) -> str:
    if sys.stdin is not sys.__stdin__:
        return _fallback_getpass(prompt)

    for char in prompt:
        msvcrt.putwch(char)

    password = ""
    while True:
        char = msvcrt.getwch()
        if char in ("\r", "\n"):
            break
        if char == "\x03":
            raise KeyboardInterrupt
        if char == "\b":
            if password:
                msvcrt.putwch("\b")
                msvcrt.putwch(" ")
                msvcrt.putwch("\b")
            password = password[:-1]
        else:
            password += char
            msvcrt.putwch("*")

    msvcrt.putwch("\r")
    msvcrt.putwch("\n")
    return password


def _fallback_getpass(prompt: str) -> str:
    warnings.warn(
        "Can not control echo on the terminal.",
        category=UserWarning,
        stacklevel=2,
    )
    print("Warning: Password input may be echoed.", file=sys.stderr)
    return input(prompt)


def _readline_with_asterisks(
    output_stream: TextIO, input_stream: TextIO, prompt: str
) -> str:
    output_stream.write(prompt)
    output_stream.flush()

    password = ""
    eof_pressed = False
    while True:
        char = input_stream.read(1)
        if char in ("\n", "\r"):
            break
        if char == "\x03":
            raise KeyboardInterrupt
        if char in ("\x7f", "\b"):
            if password:
                output_stream.write("\b \b")
                output_stream.flush()
            password = password[:-1]
        elif char == "\x04":
            if eof_pressed:
                break
            eof_pressed = True
        elif char != "\x00":
            password += char
            output_stream.write("*")
            output_stream.flush()
            eof_pressed = False
    return password



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import json
import logging
import os

import click
from mvt.common.cli_plugins import (
    IOS_CLI_PLUGIN_GROUP,
    MVT_IOS_CUSTOM_COMMANDS_ENV,
    load_cli_commands_option,
    register_cli_plugins,
)
from mvt.common.options import MutuallyExclusiveOption
from mvt.common.utils import (
    generate_hashes_from_path,
    init_logging,
    set_verbose_logging,
)
from mvt.common.help import (
    HELP_MSG_VERSION,
    HELP_MSG_DECRYPT_BACKUP,
    HELP_MSG_BACKUP_DESTINATION,
    HELP_MSG_IOS_BACKUP_PASSWORD,
    HELP_MSG_BACKUP_KEYFILE,
    HELP_MSG_HASHES,
    HELP_MSG_EXTRACT_KEY,
    HELP_MSG_IOC,
    HELP_MSG_OUTPUT,
    HELP_MSG_FAST,
    HELP_MSG_LIST_MODULES,
    HELP_MSG_LOAD_MODULE,
    HELP_MSG_MODULE,
    HELP_MSG_VERBOSE,
    HELP_MSG_VERBOSE_COMMAND,
    HELP_MSG_CHECK_FS,
    HELP_MSG_CHECK_IOCS,
    HELP_MSG_STIX2,
    HELP_MSG_CHECK_IOS_BACKUP,
    HELP_MSG_CHECK_SYSDIAGNOSE,
    HELP_MSG_DISABLE_UPDATE_CHECK,
    HELP_MSG_DISABLE_INDICATOR_UPDATE_CHECK,
)
from mvt.common.password import prompt_password

# The commands import what they run only when they are invoked. This module is
# imported at every start of mvt-ios, including by shell completion on every
# keystroke, so importing it must do no more than build the command tree: the
# forensic modules, the backup decryption and the update checks stay out of it.

init_logging()
log = logging.getLogger("mvt")

# Set this environment variable to a password if needed.
MVT_IOS_BACKUP_PASSWORD = "MVT_IOS_BACKUP_PASSWORD"
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def _get_disable_flags(ctx):
    """Helper function to safely get disable flags from context."""
    if ctx.obj is None:
        return False, False
    return (
        ctx.obj.get("disable_version_check", False),
        ctx.obj.get("disable_indicator_check", False),
    )


def _get_verbose(ctx):
    """Return whether --verbose was passed to the CLI itself."""
    return bool(ctx.obj and ctx.obj.get("verbose", False))


def _load_custom_modules(load_module):
    from mvt.common.module_loader import CustomModuleLoadError, load_custom_modules

    try:
        return load_custom_modules(load_module)
    except CustomModuleLoadError as exc:
        raise click.ClickException(str(exc)) from exc


# ==============================================================================
# Main
# ==============================================================================
@click.group(invoke_without_command=False)
@load_cli_commands_option
@click.option(
    "--disable-update-check", is_flag=True, help=HELP_MSG_DISABLE_UPDATE_CHECK
)
@click.option(
    "--disable-indicator-update-check",
    is_flag=True,
    help=HELP_MSG_DISABLE_INDICATOR_UPDATE_CHECK,
)
@click.option("--verbose", "-v", is_flag=True, help=HELP_MSG_VERBOSE)
@click.pass_context
def cli(ctx, disable_update_check, disable_indicator_update_check, verbose):
    ctx.ensure_object(dict)
    ctx.obj["disable_version_check"] = disable_update_check
    ctx.obj["disable_indicator_check"] = disable_indicator_update_check
    ctx.obj["verbose"] = verbose
    set_verbose_logging(verbose)

    from mvt.common.logo import logo

    logo(
        disable_version_check=disable_update_check,
        disable_indicator_check=disable_indicator_update_check,
    )


# ==============================================================================
# Command: version
# ==============================================================================
@cli.command("version", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_VERSION)
def version():
    return


# ==============================================================================
# Command: decrypt-backup
# ==============================================================================
@cli.command(
    "decrypt-backup", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_DECRYPT_BACKUP
)
@click.option("--destination", "-d", required=True, help=HELP_MSG_BACKUP_DESTINATION)
@click.option(
    "--password",
    "-p",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["key_file"],
    help=HELP_MSG_IOS_BACKUP_PASSWORD,
)
@click.option(
    "--key-file",
    "-k",
    cls=MutuallyExclusiveOption,
    type=click.Path(exists=True),
    mutually_exclusive=["password"],
    help=HELP_MSG_BACKUP_KEYFILE,
)
@click.option("--hashes", "-H", is_flag=True, help=HELP_MSG_HASHES)
@click.argument("BACKUP_PATH", type=click.Path(exists=True))
@click.pass_context
def decrypt_backup(ctx, destination, password, key_file, hashes, backup_path):
    from .decrypt import DecryptBackup

    backup = DecryptBackup(backup_path, destination)

    if key_file:
        if MVT_IOS_BACKUP_PASSWORD in os.environ:
            log.info(
                "Ignoring %s environment variable, using --key-file'%s' instead",
                MVT_IOS_BACKUP_PASSWORD,
                key_file,
            )

        backup.decrypt_with_key_file(key_file)
    elif password:
        log.info(
            "Your password may be visible in the process table because it "
            "was supplied on the command line!"
        )

        if MVT_IOS_BACKUP_PASSWORD in os.environ:
            log.info(
                "Ignoring %s environment variable, using --passwordargument instead",
                MVT_IOS_BACKUP_PASSWORD,
            )

        backup.decrypt_with_password(password)
    elif MVT_IOS_BACKUP_PASSWORD in os.environ:
        log.info("Using password from %s environment variable", MVT_IOS_BACKUP_PASSWORD)
        backup.decrypt_with_password(os.environ[MVT_IOS_BACKUP_PASSWORD])
    else:
        sekrit = prompt_password("Enter backup password: ")
        backup.decrypt_with_password(sekrit)

    if not backup.can_process():
        ctx.exit(1)

    backup.process_backup()

    if hashes:
        info = {"encrypted": [], "decrypted": []}
        for file in generate_hashes_from_path(backup_path, log):
            info["encrypted"].append(file)
        for file in generate_hashes_from_path(destination, log):
            info["decrypted"].append(file)
        info_path = os.path.join(destination, "info.json")
        with open(info_path, "w+", encoding="utf-8") as handle:
            json.dump(info, handle, indent=4)


# ==============================================================================
# Command: extract-key
# ==============================================================================
@cli.command(
    "extract-key", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_EXTRACT_KEY
)
@click.option("--password", "-p", help=HELP_MSG_IOS_BACKUP_PASSWORD)
@click.option(
    "--key-file",
    "-k",
    required=False,
    type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True),
    help=HELP_MSG_BACKUP_KEYFILE,
)
@click.argument("BACKUP_PATH", type=click.Path(exists=True))
def extract_key(password, key_file, backup_path):
    from .decrypt import DecryptBackup

    backup = DecryptBackup(backup_path)

    if password:
        log.info(
            "Your password may be visible in the process table because it "
            "was supplied on the command line!"
        )

        if MVT_IOS_BACKUP_PASSWORD in os.environ:
            log.info(
                "Ignoring %s environment variable, using --password argument instead",
                MVT_IOS_BACKUP_PASSWORD,
            )
    elif MVT_IOS_BACKUP_PASSWORD in os.environ:
        log.info("Using password from %s environment variable", MVT_IOS_BACKUP_PASSWORD)
        password = os.environ[MVT_IOS_BACKUP_PASSWORD]
    else:
        password = prompt_password("Enter backup password: ")

    backup.decrypt_with_password(password)
    backup.get_key()

    if key_file:
        backup.write_key(key_file)


# ==============================================================================
# Command: check-backup
# ==============================================================================
@cli.command(
    "check-backup", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_CHECK_IOS_BACKUP
)
@click.option(
    "--iocs",
    "-i",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_IOC,
)
@click.option("--output", "-o", type=click.Path(exists=False), help=HELP_MSG_OUTPUT)
@click.option("--fast", "-f", is_flag=True, help=HELP_MSG_FAST)
@click.option("--list-modules", "-l", is_flag=True, help=HELP_MSG_LIST_MODULES)
@click.option("--module", "-m", help=HELP_MSG_MODULE)
@click.option(
    "--load-module",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_LOAD_MODULE,
)
@click.option("--hashes", "-H", is_flag=True, help=HELP_MSG_HASHES)
@click.option("--verbose", "-v", is_flag=True, help=HELP_MSG_VERBOSE_COMMAND)
@click.argument("BACKUP_PATH", type=click.Path(exists=True))
@click.pass_context
def check_backup(
    ctx,
    iocs,
    output,
    fast,
    list_modules,
    module,
    load_module,
    hashes,
    verbose,
    backup_path,
):
    from .cmd_check_backup import CmdIOSCheckBackup

    set_verbose_logging(verbose or _get_verbose(ctx))
    module_options = {"fast_mode": fast}
    custom_modules = _load_custom_modules(load_module)

    cmd = CmdIOSCheckBackup(
        target_path=backup_path,
        results_path=output,
        ioc_files=iocs,
        module_name=module,
        module_options=module_options,
        hashes=hashes,
        disable_version_check=_get_disable_flags(ctx)[0],
        disable_indicator_check=_get_disable_flags(ctx)[1],
        custom_modules=custom_modules,
    )

    if list_modules:
        cmd.list_modules()
        return

    if not cmd.resolve_backup_path():
        ctx.exit(1)

    log.info("Checking iTunes backup located at: %s", cmd.target_path)

    cmd.run()
    cmd.show_alerts_brief()
    cmd.show_support_message()


# ==============================================================================
# Command: check-fs
# ==============================================================================
@cli.command("check-fs", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_CHECK_FS)
@click.option(
    "--iocs",
    "-i",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_IOC,
)
@click.option("--output", "-o", type=click.Path(exists=False), help=HELP_MSG_OUTPUT)
@click.option("--fast", "-f", is_flag=True, help=HELP_MSG_FAST)
@click.option("--list-modules", "-l", is_flag=True, help=HELP_MSG_LIST_MODULES)
@click.option("--module", "-m", help=HELP_MSG_MODULE)
@click.option(
    "--load-module",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_LOAD_MODULE,
)
@click.option("--hashes", "-H", is_flag=True, help=HELP_MSG_HASHES)
@click.option("--verbose", "-v", is_flag=True, help=HELP_MSG_VERBOSE_COMMAND)
@click.argument("DUMP_PATH", type=click.Path(exists=True))
@click.pass_context
def check_fs(
    ctx,
    iocs,
    output,
    fast,
    list_modules,
    module,
    load_module,
    hashes,
    verbose,
    dump_path,
):
    from .cmd_check_fs import CmdIOSCheckFS

    set_verbose_logging(verbose or _get_verbose(ctx))
    module_options = {"fast_mode": fast}
    custom_modules = _load_custom_modules(load_module)

    cmd = CmdIOSCheckFS(
        target_path=dump_path,
        results_path=output,
        ioc_files=iocs,
        module_name=module,
        module_options=module_options,
        hashes=hashes,
        disable_version_check=_get_disable_flags(ctx)[0],
        disable_indicator_check=_get_disable_flags(ctx)[1],
        custom_modules=custom_modules,
    )

    if list_modules:
        cmd.list_modules()
        return

    log.info("Checking iOS filesystem located at: %s", dump_path)

    cmd.run()
    cmd.show_alerts_brief()
    cmd.show_support_message()


# ==============================================================================
# Command: check-sysdiagnose
# ==============================================================================
@cli.command(
    "check-sysdiagnose",
    context_settings=CONTEXT_SETTINGS,
    help=HELP_MSG_CHECK_SYSDIAGNOSE,
)
@click.option(
    "--iocs",
    "-i",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_IOC,
)
@click.option("--output", "-o", type=click.Path(exists=False), help=HELP_MSG_OUTPUT)
@click.option("--list-modules", "-l", is_flag=True, help=HELP_MSG_LIST_MODULES)
@click.option("--module", "-m", help=HELP_MSG_MODULE)
@click.option(
    "--load-module",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_LOAD_MODULE,
)
@click.option("--hashes", "-H", is_flag=True, help=HELP_MSG_HASHES)
@click.option("--verbose", "-v", is_flag=True, help=HELP_MSG_VERBOSE_COMMAND)
@click.argument("SYSDIAGNOSE_PATH", type=click.Path(exists=True))
@click.pass_context
def check_sysdiagnose(
    ctx,
    iocs,
    output,
    list_modules,
    module,
    load_module,
    hashes,
    verbose,
    sysdiagnose_path,
):
    from .cmd_check_sysdiagnose import CmdIOSCheckSysdiagnose

    set_verbose_logging(verbose or _get_verbose(ctx))
    custom_modules = _load_custom_modules(load_module)
    cmd = CmdIOSCheckSysdiagnose(
        target_path=sysdiagnose_path,
        results_path=output,
        ioc_files=iocs,
        module_name=module,
        hashes=hashes,
        disable_version_check=_get_disable_flags(ctx)[0],
        disable_indicator_check=_get_disable_flags(ctx)[1],
        custom_modules=custom_modules,
    )

    if not cmd._available_modules():
        raise click.ClickException(
            "No custom modules support mvt-ios check-sysdiagnose. "
            "Load a module that declares supported_commands = "
            "((\"ios\", \"check-sysdiagnose\"),)."
        )

    if list_modules:
        cmd.list_modules()
        return

    log.info("Checking iOS sysdiagnose at path: %s", sysdiagnose_path)
    cmd.run()
    cmd.show_alerts_brief()
    cmd.show_support_message()


# ==============================================================================
# Command: check-iocs
# ==============================================================================
@cli.command("check-iocs", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_CHECK_IOCS)
@click.option(
    "--iocs",
    "-i",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_IOC,
)
@click.option("--list-modules", "-l", is_flag=True, help=HELP_MSG_LIST_MODULES)
@click.option("--module", "-m", help=HELP_MSG_MODULE)
@click.option(
    "--load-module",
    type=click.Path(exists=True),
    multiple=True,
    default=[],
    help=HELP_MSG_LOAD_MODULE,
)
@click.argument("FOLDER", type=click.Path(exists=True))
@click.pass_context
def check_iocs(ctx, iocs, list_modules, module, load_module, folder):
    from mvt.common.cmd_check_iocs import CmdCheckIOCS

    from .command_modules import IOS_CHECK_IOCS_MODULES

    custom_modules = _load_custom_modules(load_module)
    cmd = CmdCheckIOCS(
        target_path=folder,
        ioc_files=iocs,
        module_name=module,
        disable_version_check=_get_disable_flags(ctx)[0],
        disable_indicator_check=_get_disable_flags(ctx)[1],
        custom_modules=custom_modules,
        platform="ios",
    )
    cmd.modules = IOS_CHECK_IOCS_MODULES

    if list_modules:
        cmd.list_modules()
        return

    cmd.run()
    cmd.show_alerts_brief()
    cmd.show_support_message()


# ==============================================================================
# Command: download-iocs
# ==============================================================================
@cli.command("download-iocs", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_STIX2)
def download_iocs():
    from mvt.common.updates import IndicatorsUpdates

    ioc_updates = IndicatorsUpdates()
    ioc_updates.update()


# ==============================================================================
# Entry point of the mvt-ios console script
# ==============================================================================
def main() -> None:
    """Register the external commands and run the mvt-ios CLI.

    External commands are registered here rather than when this module is
    imported, so that importing MVT never runs third-party code and a plugin
    importing from MVT cannot re-enter a module that is still initializing.
    """
    register_cli_plugins(
        cli,
        entry_point_group=IOS_CLI_PLUGIN_GROUP,
        environment_variable=MVT_IOS_CUSTOM_COMMANDS_ENV,
    )
    cli()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_chrometime_to_datetime, convert_datetime_to_iso

from ..base import IOSExtraction

CHROME_FAVICON_BACKUP_IDS = ["55680ab883d0fdcffd94f959b1632e5fbbb18c5b"]
CHROME_FAVICON_ROOT_PATHS = [
    "private/var/mobile/Containers/Data/Application/*/Library/Application Support/Google/Chrome/Default/Favicons",
]


class ChromeFavicon(IOSExtraction):
    """This module extracts all Chrome favicon records."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "new_favicon",
            "data": f"{record['icon_url']} from {record['url']}",
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_url(result["url"])
            if not ioc_match:
                ioc_match = self.indicators.check_url(result["icon_url"])

            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )
                continue

    def run(self) -> None:
        self._find_ios_database(
            backup_ids=CHROME_FAVICON_BACKUP_IDS, root_paths=CHROME_FAVICON_ROOT_PATHS
        )
        self.log.info("Found Chrome favicon cache database at path: %s", self.file_path)

        if not self.file_path:
            return
        conn = self._open_sqlite_db(self.file_path)

        # Fetch icon cache
        cur = conn.cursor()
        cur.execute(
            """
            SELECT
                icon_mapping.page_url,
                favicons.url,
                favicon_bitmaps.last_updated,
                favicon_bitmaps.last_requested
            FROM icon_mapping
            JOIN favicon_bitmaps ON icon_mapping.icon_id = favicon_bitmaps.icon_id
            JOIN favicons ON icon_mapping.icon_id = favicons.id
            ORDER BY icon_mapping.id;
        """
        )

        records = []
        for row in cur:
            last_timestamp = int(row[2]) or int(row[3])
            records.append(
                {
                    "url": row[0],
                    "icon_url": row[1],
                    "timestamp": last_timestamp,
                    "isodate": convert_datetime_to_iso(
                        convert_chrometime_to_datetime(last_timestamp)
                    ),
                }
            )

        cur.close()
        conn.close()

        self.log.info("Extracted a total of %d favicon records", len(records))
        self.results = sorted(records, key=lambda row: row["isodate"])



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
from typing import Optional

from mvt.common.command import Command
from mvt.common.module import MVTModule
from mvt.common.module_loader import get_module_logger
from mvt.common.utils import exec_or_profile

log = logging.getLogger(__name__)


class CmdCheckIOCS(Command):
    def __init__(
        self,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        ioc_files: Optional[list] = None,
        module_name: Optional[str] = None,
        serial: Optional[str] = None,
        module_options: Optional[dict] = None,
        hashes: Optional[bool] = False,
        sub_command: Optional[bool] = False,
        disable_version_check: bool = False,
        disable_indicator_check: bool = False,
        custom_modules: Optional[list[type[MVTModule]]] = None,
        platform: str = "",
    ) -> None:
        super().__init__(
            target_path=target_path,
            results_path=results_path,
            ioc_files=ioc_files,
            module_name=module_name,
            serial=serial,
            module_options=module_options,
            hashes=hashes,
            sub_command=sub_command,
            log=log,
            disable_version_check=disable_version_check,
            disable_indicator_check=disable_indicator_check,
            custom_modules=custom_modules,
        )

        self.platform = platform
        self.name = "check-iocs"

    def run(self) -> None:
        assert self.target_path is not None
        all_modules = []
        for entry in self._available_modules():
            if entry not in all_modules:
                all_modules.append(entry)

        log.info("Checking stored results against provided indicators...")

        total_detections = 0
        for file_name in os.listdir(self.target_path):
            name_only, _ = os.path.splitext(file_name)
            file_path = os.path.join(self.target_path, file_name)

            for iocs_module in all_modules:
                if self.module_name and iocs_module.__name__ != self.module_name:
                    continue

                if iocs_module.get_slug() != name_only:
                    continue

                log.info(
                    'Loading results from "%s" with module %s',
                    file_name,
                    iocs_module.__name__,
                )

                m = iocs_module.from_json(
                    file_path, log=get_module_logger(iocs_module)
                )
                if not m:
                    log.warning("No result from this module, skipping it")
                    continue

                if self.iocs.total_ioc_count > 0:
                    m.indicators = self.iocs
                    m.indicators.log = m.log

                try:
                    exec_or_profile("m.check_indicators()", globals(), locals())
                except NotImplementedError:
                    continue
                else:
                    total_detections += len(m.alertstore.alerts)

        if total_detections > 0:
            log.warning(
                "The check of the results produced %d detections!", total_detections
            )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2026 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import click

from mvt.common.cli_plugins import (
    MVT_CUSTOM_COMMANDS_ENV,
    NEUTRAL_CLI_PLUGIN_GROUP,
    load_cli_commands_option,
    register_cli_plugins,
)
from mvt.common.cmd_plugins import plugins
from mvt.common.completion import completion
from mvt.common.help import (
    HELP_MSG_DISABLE_INDICATOR_UPDATE_CHECK,
    HELP_MSG_DISABLE_UPDATE_CHECK,
    HELP_MSG_STIX2,
    HELP_MSG_VERBOSE,
    HELP_MSG_VERSION,
)
from mvt.common.logo import logo
from mvt.common.updates import IndicatorsUpdates
from mvt.common.utils import init_logging, set_verbose_logging

init_logging()

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


# ==============================================================================
# Main
# ==============================================================================
@click.group(invoke_without_command=True)
@load_cli_commands_option
@click.option(
    "--disable-update-check", is_flag=True, help=HELP_MSG_DISABLE_UPDATE_CHECK
)
@click.option(
    "--disable-indicator-update-check",
    is_flag=True,
    help=HELP_MSG_DISABLE_INDICATOR_UPDATE_CHECK,
)
@click.option("--verbose", "-v", is_flag=True, help=HELP_MSG_VERBOSE)
@click.pass_context
def cli(ctx, disable_update_check, disable_indicator_update_check, verbose):
    """Mobile Verification Toolkit.

    The 'mvt-ios' and 'mvt-android' CLI commands are used to perform
    forensic analysis on IOS and Android devices.
    """
    ctx.ensure_object(dict)
    ctx.obj["disable_version_check"] = disable_update_check
    ctx.obj["disable_indicator_check"] = disable_indicator_update_check
    ctx.obj["verbose"] = verbose
    set_verbose_logging(verbose)
    if ctx.invoked_subcommand != "completion":
        logo(
            disable_version_check=disable_update_check,
            disable_indicator_check=disable_indicator_update_check,
        )
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


# ==============================================================================
# Command: download-iocs
# ==============================================================================
@cli.command("download-iocs", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_STIX2)
def download_iocs():
    ioc_updates = IndicatorsUpdates()
    ioc_updates.update()


# ==============================================================================
# Command: completion
# ==============================================================================
cli.add_command(completion)


# ==============================================================================
# Command: version
# ==============================================================================
@cli.command("version", context_settings=CONTEXT_SETTINGS, help=HELP_MSG_VERSION)
def version():
    return


# The plugins command is registered as a built-in command, before any external
# command, so that an installed package can never replace it.
cli.add_command(plugins)


# ==============================================================================
# Entry point of the mvt console script
# ==============================================================================
def main() -> None:
    """Register the external commands and run the mvt CLI.

    External commands are registered here rather than when this module is
    imported, so that importing MVT never runs third-party code and a plugin
    importing from MVT cannot re-enter a module that is still initializing.
    """
    register_cli_plugins(
        cli,
        entry_point_group=NEUTRAL_CLI_PLUGIN_GROUP,
        environment_variable=MVT_CUSTOM_COMMANDS_ENV,
    )
    cli()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import datetime
import logging
import os
from typing import Optional

from mvt.android.artifacts.file_timestamps import FileTimestampsArtifact
from mvt.common.module_types import ModuleResults
from mvt.common.utils import convert_datetime_to_iso

from .base import AndroidQFModule


class AQFLogTimestamps(FileTimestampsArtifact, AndroidQFModule):
    """This module creates timeline for log files extracted by AQF."""

    slug = "aqf_log_timestamps"

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def _get_file_modification_time(self, file_path: str) -> datetime.datetime:
        if self.archive:
            file_timetuple = self.archive.getinfo(file_path).date_time
            return datetime.datetime(*file_timetuple)
        else:
            if not self.parent_path:
                raise ValueError("parent_path is not set")
            file_stat = os.stat(os.path.join(self.parent_path, file_path))
            return datetime.datetime.fromtimestamp(file_stat.st_mtime)

    def run(self) -> None:
        filesystem_files = self._get_files_by_pattern("*/logs/*")

        self.results = []
        for file in filesystem_files:
            # Only the modification time is available in the zip file metadata.
            # The timezone is the local timezone of the machine the phone.
            modification_time = self._get_file_modification_time(file)
            self.results.append(
                {
                    "path": file,
                    "modified_time": convert_datetime_to_iso(modification_time),
                }
            )

        self.log.info(
            "Extracted a total of %d filesystem timestamps from AndroidQF logs directory.",
            len(self.results),
        )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
from datetime import datetime
from typing import Optional

from mvt.ios.versions import find_version_by_build
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleSerializedResult,
    ModuleResults,
)

from ..base import IOSExtraction
from .analytics import Analytics


class AnalyticsIOSVersions(IOSExtraction):
    """This module leverages the Analytics module in order to extract
    a timeline of build numbers from the private/var/Keychains/Analytics/*.db
    files."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "analytics_ios_version",
            "data": f"Seen iOS version {record['version']} ({record['build']})",
        }

    def run(self):
        anl = Analytics(target_path=self.target_path, log=self.log)
        anl.process_analytics_dbs()

        dt_format = "%Y-%m-%d %H:%M:%S.%f"

        builds = {}
        for result in anl.results:
            build = result.get("build")
            if not build:
                continue

            isodate = result.get("isodate", None)
            if not isodate:
                continue

            if build not in builds.keys():
                builds[build] = isodate
                continue

            result_dt = datetime.strptime(isodate, dt_format)
            cur_dt = datetime.strptime(builds[build], dt_format)

            if result_dt < cur_dt:
                builds[build] = isodate

        for build, isodate in builds.items():
            version = find_version_by_build(build)

            self.results.append(
                {
                    "isodate": isodate,
                    "build": build,
                    "version": version,
                }
            )

        self.results = sorted(self.results, key=lambda entry: entry["isodate"])
        for result in self.results:
            self.log.info(
                "iOS version %s (%s) first appeared on %s",
                result["version"],
                result["build"],
                result["isodate"],
            )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
import sqlite3
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_unix_to_iso

from ..base import IOSExtraction

WEBKIT_RESOURCELOADSTATICS_BACKUP_RELPATH = (
    "Library/WebKit/WebsiteData/ResourceLoadStatistics/observations.db"  # pylint: disable=line-too-long
)
WEBKIT_RESOURCELOADSTATICS_ROOT_PATHS = [
    "private/var/mobile/Containers/Data/Application/*/Library/WebKit/WebsiteData/ResourceLoadStatistics/observations.db",  # pylint: disable=line-too-long
    "private/var/mobile/Containers/Data/Application/*/SystemData/com.apple.SafariViewService/Library/WebKit/WebsiteData/observations.db",  # pylint: disable=line-too-long
]


class WebkitResourceLoadStatistics(IOSExtraction):
    """This module extracts records from WebKit ResourceLoadStatistics
    observations.db."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

        self.results = [] if not results else results

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        msg = f"Webkit resource loaded from {record['registrable_domain']}"
        if record["domain"] != "":
            msg += f" by app in domain {record['domain']}"
        return {
            "timestamp": record["last_seen_isodate"],
            "module": self.__class__.__name__,
            "event": "visit",
            "data": msg,
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            ioc_match = self.indicators.check_url(result["registrable_domain"])
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

    def _process_observations_db(self, db_path: str, domain: str, path: str) -> None:
        self.log.info(
            "Found WebKit ResourceLoadStatistics observations.db file at path %s",
            db_path,
        )

        self._recover_sqlite_db_if_needed(db_path)

        conn = self._open_sqlite_db(db_path)
        cur = conn.cursor()

        try:
            try:
                cur.execute("PRAGMA table_info(ObservedDomains);")
                available_columns = {row[1] for row in cur}
                required_columns = [
                    "domainID",
                    "registrableDomain",
                    "lastSeen",
                    "hadUserInteraction",
                ]
                if not set(required_columns).issubset(available_columns):
                    return

                optional_columns = [
                    column
                    for column in [
                        "mostRecentUserInteractionTime",
                        "mostRecentWebPushInteractionTime",
                    ]
                    if column in available_columns
                ]
                selected_columns = required_columns + optional_columns
                cur.execute(
                    f"SELECT {', '.join(selected_columns)} FROM ObservedDomains;"
                )
            except sqlite3.OperationalError:
                return

            for row in cur:
                result = {
                    "domain_id": row[0],
                    "registrable_domain": row[1],
                    "last_seen": row[2],
                    "had_user_interaction": bool(row[3]),
                    "last_seen_isodate": convert_unix_to_iso(row[2]),
                    "domain": domain,
                    "path": path,
                }
                for index, column in enumerate(optional_columns, start=4):
                    field = {
                        "mostRecentUserInteractionTime": (
                            "most_recent_user_interaction_time"
                        ),
                        "mostRecentWebPushInteractionTime": (
                            "most_recent_web_push_interaction_time"
                        ),
                    }[column]
                    timestamp = row[index]
                    result[field] = timestamp
                    if timestamp is not None and timestamp >= 0:
                        result[f"{field}_isodate"] = convert_unix_to_iso(timestamp)
                self.results.append(result)
        finally:
            cur.close()
            conn.close()

        if len(self.results) > 0:
            self.log.info(
                "Extracted a total of %d records from %s", len(self.results), db_path
            )

    def run(self) -> None:
        if self.is_backup:
            try:
                for backup_file in self._get_backup_files_from_manifest(
                    relative_path=WEBKIT_RESOURCELOADSTATICS_BACKUP_RELPATH
                ):
                    db_path = self._get_backup_file_from_id(backup_file["file_id"])

                    if db_path:
                        self._process_observations_db(
                            db_path=db_path,
                            domain=backup_file["domain"],
                            path=WEBKIT_RESOURCELOADSTATICS_BACKUP_RELPATH,
                        )
            except Exception as exc:
                self.log.info("Unable to find WebKit observations.db: %s", exc)
        elif self.is_fs_dump:
            for db_path in self._get_fs_files_from_patterns(
                WEBKIT_RESOURCELOADSTATICS_ROOT_PATHS
            ):
                db_rel_path = os.path.relpath(db_path, self.target_path)
                self._process_observations_db(
                    db_path=db_path, domain="", path=db_rel_path
                )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import re

from .artifact import AndroidArtifact

ANDROID_DANGEROUS_SETTINGS = [
    {
        "description": "disabled Google Play Services apps verification",
        "key": "verifier_verify_adb_installs",
        "safe_value": "1",
    },
    {
        "description": "disabled Google Play Protect",
        "key": "package_verifier_enable",
        "safe_value": "1",
    },
    {
        "description": "disabled APK package verification",
        "key": "package_verifier_state",
        "safe_value": "1",
    },
    {
        "description": "disabled Google Play Protect",
        "key": "package_verifier_user_consent",
        "safe_value": "1",
    },
    {
        "description": "disabled Google Play Protect",
        "key": "upload_apk_enable",
        "safe_value": "1",
    },
    {
        "description": "disabled confirmation of adb apps installation",
        "key": "adb_install_need_confirm",
        "safe_value": "1",
    },
    {
        "description": "disabled sharing of security reports",
        "key": "send_security_reports",
        "safe_value": "1",
    },
    {
        "description": "disabled sharing of crash logs with manufacturer",
        "key": "samsung_errorlog_agree",
        "safe_value": "1",
    },
    {
        "description": "disabled applications errors reports",
        "key": "send_action_app_error",
        "safe_value": "1",
    },
    {
        "description": "enabled accessibility services",
        "key": "accessibility_enabled",
        "safe_value": "0",
    },
]


class Settings(AndroidArtifact):
    def parse(self, content: str) -> None:
        self.results: dict[str, dict[str, str]] = {}
        namespace: str | None = None
        for line in content.splitlines():
            heading = re.match(
                r"^(CONFIG|GLOBAL|SECURE|SYSTEM) SETTINGS \(user (\d+)\)$",
                line.strip(),
            )
            if heading:
                namespace = f"{heading.group(1).lower()}:user_{heading.group(2)}"
                self.results[namespace] = {}
                continue
            if namespace is None or not line.startswith("_id:"):
                continue
            setting = re.match(
                r"^_id:\S+\s+name:(.*?)\s+pkg:.*?\s+value:(.*?)"
                r"(?:\s+default:.*\s+defaultSystemSet:(?:true|false))?$",
                line,
            )
            if setting:
                self.results[namespace][setting.group(1)] = setting.group(2)

    def check_indicators(self) -> None:
        for namespace, settings in self.results.items():
            for key, value in settings.items():
                for danger in ANDROID_DANGEROUS_SETTINGS:
                    # Check if one of the dangerous settings is using an unsafe
                    # value (different than the one specified).
                    if danger["key"] == key and danger["safe_value"] != value:
                        self.alertstore.medium(
                            f'Found suspicious "{namespace}" setting "{key} = {value}" ({danger["description"]})',
                            "",
                            {
                                "namespace": namespace,
                                "key": key,
                                "value": value,
                                "description": danger["description"],
                            },
                        )
                        break



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import re

from .artifact import AndroidArtifact


class DumpsysDBInfoArtifact(AndroidArtifact):
    """
    Parser for dumpsys DBInfo service
    """

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            path = result.get("path", "")
            for part in path.split("/"):
                ioc_match = self.indicators.check_app_id(part)
                if ioc_match:
                    self.alertstore.critical(
                        ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                    )
                    continue

    def parse(self, output: str) -> None:
        rxp = re.compile(
            r"^\s*\d+:\s*\[((?:\d{4}-)?\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3})\]\s*"
            r"(?:\[Pid:\((\d+)\)\])?\s*([\w-]+) took (\d+)ms - ([^,]+),"
            r"\s*sql=\"(.*)\"(?:, path=(.*))?$"
        )

        pool: str | None = None
        connection_number: int | None = None
        is_primary: bool | None = None
        in_operations = False
        for line in output.splitlines():
            if line.startswith("Connection pool for "):
                pool = line.replace("Connection pool for ", "").rstrip(":")
                in_operations = False

            connection_match = re.match(r"\s+Connection #(\d+):", line)
            if connection_match:
                connection_number = int(connection_match.group(1))
                is_primary = None

            if line.strip().startswith("isPrimaryConnection:"):
                is_primary = line.strip().split(":", 1)[1].strip() == "true"

            if not pool:
                continue

            if line.strip() == "Most recently executed operations:":
                in_operations = True
                continue

            if not in_operations:
                continue

            if not line.startswith("        "):
                in_operations = False
                continue

            match = rxp.match(line)
            if not match:
                continue

            result = {
                "timestamp": match.group(1),
                "pid": int(match.group(2)) if match.group(2) else None,
                "action": match.group(3),
                "duration_ms": int(match.group(4)),
                "status": match.group(5),
                "sql": match.group(6),
                "path": match.group(7) or pool,
                "pool_path": pool,
                "connection_number": connection_number,
                "is_primary": is_primary,
            }
            self.results.append(result)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import os
from typing import Optional

from mvt.common.utils import convert_unix_to_iso
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleSerializedResult,
    ModuleResults,
)

from ..base import IOSExtraction


class Filesystem(IOSExtraction):
    """This module extracts creation and modification date of files from a
    full file-system dump.
    """

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["modified"],
            "module": self.__class__.__name__,
            "event": "entry_modified",
            "data": record["path"],
        }

    def check_indicators(self) -> None:
        if not self.indicators:
            return

        for result in self.results:
            if "path" not in result:
                continue

            ioc_match = self.indicators.check_file_path(result["path"])
            if ioc_match:
                self.alertstore.high(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

            # If we are instructed to run fast, we skip the rest.
            if self.module_options.get("fast_mode", None):
                continue

            ioc_match = self.indicators.check_file_path_process(result["path"])
            if ioc_match:
                self.alertstore.high(
                    ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                )

    def run(self) -> None:
        if not self.target_path:
            self.log.error("No filesystem dump path provided")
            return

        for root, dirs, files in os.walk(self.target_path):
            for dir_name in dirs:
                try:
                    dir_path = os.path.join(root, dir_name)
                    result = {
                        "path": os.path.relpath(dir_path, self.target_path),
                        "modified": convert_unix_to_iso(os.stat(dir_path).st_mtime),
                    }
                except Exception:
                    continue
                else:
                    self.results.append(result)

            for file_name in files:
                try:
                    file_path = os.path.join(root, file_name)
                    result = {
                        "path": os.path.relpath(file_path, self.target_path),
                        "modified": convert_unix_to_iso(os.stat(file_path).st_mtime),
                    }
                except Exception:
                    continue
                else:
                    self.results.append(result)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2025 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import csv
import inspect
import logging
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from .log import CRITICAL_ALERT, HIGH_ALERT, INFO_ALERT, LOW_ALERT, MEDIUM_ALERT
from .module_types import ModuleAtomicResult


class AlertLevel(Enum):
    INFORMATIONAL = 0
    LOW = 10
    MEDIUM = 20
    HIGH = 30
    CRITICAL = 40


@dataclass
class Alert:
    level: AlertLevel
    module: str
    message: str
    event_time: str
    event: ModuleAtomicResult
    matched_indicator: Optional[Any] = None


class AlertStore:
    def __init__(self, log: Optional[logging.Logger] = None) -> None:
        self.__alerts: List[Alert] = []
        self.__log = log

    def _get_calling_module(self) -> str:
        """
        Automatically detect the calling MVT module and return its slug.

        Walks up the call stack to find the first frame that belongs to an MVT module
        (artifact or extraction module) and extracts its slug.

        :return: Module slug string
        """
        frame = inspect.currentframe()
        try:
            # Walk up the call stack
            while frame is not None:
                frame = frame.f_back
                if frame is None:
                    break

                # Get the 'self' object from the frame's local variables
                frame_locals = frame.f_locals
                if "self" in frame_locals:
                    obj = frame_locals["self"]
                    # Check if it has a get_slug method (MVT modules have this)
                    if hasattr(obj, "get_slug") and callable(obj.get_slug):
                        try:
                            return str(obj.get_slug())
                        except Exception:
                            pass

            # Fallback: return "unknown" if we can't find the module
            return "unknown"
        finally:
            del frame

    @property
    def alerts(self) -> List[Alert]:
        return self.__alerts

    def add(self, alert: Alert) -> None:
        self.__alerts.append(alert)
        self.log(alert)

    def extend(self, alerts: List[Alert]) -> None:
        for alert in alerts:
            self.add(alert)

    def info(
        self,
        message: str,
        event_time: str,
        event: ModuleAtomicResult,
        matched_indicator: Optional[Any] = None,
    ):
        self.add(
            Alert(
                level=AlertLevel.INFORMATIONAL,
                module=self._get_calling_module(),
                message=message,
                event_time=event_time,
                event=event,
                matched_indicator=matched_indicator,
            )
        )

    def low(
        self,
        message: str,
        event_time: str,
        event: ModuleAtomicResult,
        matched_indicator: Optional[Any] = None,
    ):
        self.add(
            Alert(
                level=AlertLevel.LOW,
                module=self._get_calling_module(),
                message=message,
                event_time=event_time,
                event=event,
                matched_indicator=matched_indicator,
            )
        )

    def medium(
        self,
        message: str,
        event_time: str,
        event: ModuleAtomicResult,
        matched_indicator: Optional[Any] = None,
    ):
        self.add(
            Alert(
                level=AlertLevel.MEDIUM,
                module=self._get_calling_module(),
                message=message,
                event_time=event_time,
                event=event,
                matched_indicator=matched_indicator,
            )
        )

    def high(
        self,
        message: str,
        event_time: str,
        event: ModuleAtomicResult,
        matched_indicator: Optional[Any] = None,
    ):
        self.add(
            Alert(
                level=AlertLevel.HIGH,
                module=self._get_calling_module(),
                message=message,
                event_time=event_time,
                event=event,
                matched_indicator=matched_indicator,
            )
        )

    def critical(
        self,
        message: str,
        event_time: str,
        event: ModuleAtomicResult,
        matched_indicator: Optional[Any] = None,
    ):
        self.add(
            Alert(
                level=AlertLevel.CRITICAL,
                module=self._get_calling_module(),
                message=message,
                event_time=event_time,
                event=event,
                matched_indicator=matched_indicator,
            )
        )

    def log(self, alert: Alert) -> None:
        if not self.__log:
            return

        if not alert.message:
            return

        if alert.level == AlertLevel.INFORMATIONAL:
            self.__log.log(INFO_ALERT, alert.message)
        elif alert.level == AlertLevel.LOW:
            self.__log.log(LOW_ALERT, alert.message)
        elif alert.level == AlertLevel.MEDIUM:
            self.__log.log(MEDIUM_ALERT, alert.message)
        elif alert.level == AlertLevel.HIGH:
            self.__log.log(HIGH_ALERT, alert.message)
        elif alert.level == AlertLevel.CRITICAL:
            self.__log.log(CRITICAL_ALERT, alert.message)

    def log_latest(self) -> None:
        self.log(self.__alerts[-1])

    def count(self, level: AlertLevel) -> int:
        count = 0
        for alert in self.__alerts:
            if alert.level == level:
                count += 1

        return count

    def as_json(self) -> List[Dict[str, Any]]:
        alerts = []
        for alert in self.__alerts:
            alert_dict = asdict(alert)
            # This is required because an Enum is not JSON serializable.
            alert_dict["level"] = alert.level.name
            if isinstance(alert_dict.get("event"), dict):
                event_matched_indicator = alert_dict["event"].pop(
                    "matched_indicator", None
                )
                if alert_dict["matched_indicator"] is None:
                    alert_dict["matched_indicator"] = event_matched_indicator

            alerts.append(alert_dict)

        return alerts

    def save_timeline(self, timeline_path: str) -> None:
        with open(timeline_path, "w", encoding="utf-8") as handle:
            csvoutput = csv.writer(
                handle,
                delimiter=",",
                quotechar='"',
                quoting=csv.QUOTE_ALL,
                escapechar="\\",
            )
            csvoutput.writerow(["Event Time", "Module", "Message", "Event"])

            timed_alerts = []
            for alert in self.alerts:
                if not alert.event_time:
                    continue

                timed_alerts.append(asdict(alert))

            for event in sorted(
                timed_alerts,
                key=lambda x: x["event_time"] if x["event_time"] is not None else "",
            ):
                csvoutput.writerow(
                    [
                        event.get("event_time"),
                        event.get("module"),
                        event.get("message"),
                        event.get("event"),
                    ]
                )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

from mvt.common.module import MVTModule

from .analytics import Analytics
from .analytics_ios_versions import AnalyticsIOSVersions
from .cache_files import CacheFiles
from .filesystem import Filesystem
from .net_netusage import Netusage
from .safari_favicon import SafariFavicon
from .shutdownlog import ShutdownLog
from .version_history import IOSVersionHistory
from .webkit_indexeddb import WebkitIndexedDB
from .webkit_localstorage import WebkitLocalStorage
from .webkit_safariviewservice import WebkitSafariViewService

FS_MODULES: list[type[MVTModule]] = [
    CacheFiles,
    Filesystem,
    Netusage,
    Analytics,
    AnalyticsIOSVersions,
    SafariFavicon,
    ShutdownLog,
    IOSVersionHistory,
    WebkitIndexedDB,
    WebkitLocalStorage,
    WebkitSafariViewService,
]



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import glob
import json
import logging
import os
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Dict, Iterator, List, Optional, Sequence

import ahocorasick
from appdirs import user_data_dir

from .config import settings
from .url import URL

MVT_DATA_FOLDER = user_data_dir("mvt")
MVT_INDICATORS_FOLDER = os.path.join(MVT_DATA_FOLDER, "indicators")

logger = logging.getLogger(__name__)

URL_CHECK_MAX_WORKERS = 20


@dataclass
class Indicator:
    value: str
    type: str
    name: str
    stix2_file_name: str


@dataclass
class IndicatorMatch:
    ioc: Indicator
    message: str


class Indicators:
    """This class is used to parse indicators from a STIX2 file and provide
    functions to compare extracted artifacts to the indicators.
    """

    def __init__(self, log=logger) -> None:
        self.log = log
        self.ioc_collections: List[Dict[str, Any]] = []
        self.total_ioc_count = 0
        self.resolved_urls: Dict[str, str] = {}

    def _load_downloaded_indicators(self) -> None:
        if not os.path.isdir(MVT_INDICATORS_FOLDER):
            return

        for ioc_file_name in os.listdir(MVT_INDICATORS_FOLDER):
            if ioc_file_name.lower().endswith(".stix2"):
                self.parse_stix2(os.path.join(MVT_INDICATORS_FOLDER, ioc_file_name))

    def _check_stix2_env_variable(self) -> None:
        """
        Checks if MVT_STIX2 setting or environment variable contains path to a STIX file. Also recursively searches through dirs in MVT_STIX2
        """
        if not settings.STIX2:
            return

        paths = settings.STIX2.split(":")
        for path in paths:
            if os.path.isfile(path) and path.lower().endswith(".stix2"):
                self.parse_stix2(path)
            elif os.path.isdir(path):
                for file in glob.glob(os.path.join(path, "**", "*.stix2"), recursive=True):
                    self.parse_stix2(file)
            else:
                self.log.error(
                    "Path specified with env MVT_STIX2 is not a valid path: '%s'", path
                )

    def _new_collection(
        self,
        cid: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        file_name: Optional[str] = None,
        file_path: Optional[str] = None,
    ) -> dict:
        return {
            "id": cid,
            "name": name,
            "description": description,
            "stix2_file_name": file_name,
            "stix2_file_path": file_path,
            "domains": [],
            "processes": [],
            "emails": [],
            "file_names": [],
            "file_paths": [],
            "files_md5": [],
            "files_sha1": [],
            "files_sha256": [],
            "app_cert_hashes": [],
            "app_ids": [],
            "ios_profile_ids": [],
            "android_property_names": [],
            "urls": [],
            "count": 0,
        }

    def _add_indicator(self, ioc: str, ioc_coll: dict, ioc_coll_list: list) -> None:
        ioc = ioc.replace("'", "").strip()
        if ioc not in ioc_coll_list:
            ioc_coll_list.append(ioc)
            ioc_coll["count"] += 1
            self.total_ioc_count += 1

    def _process_indicator(self, indicator: dict, collection: dict) -> None:
        key, value = indicator.get("pattern", "").strip("[]").split("=")
        key = key.strip()

        # Normalize hash algorithm keys so that both the STIX2-spec-compliant
        # form (e.g. file:hashes.'SHA-256', which requires quotes around
        # algorithm names that contain hyphens) and the non-standard lowercase
        # form (e.g. file:hashes.sha256) are accepted.  Strip single quotes and
        # hyphens from the algorithm name only, then lowercase it.
        for sep in ("hashes.", "cert."):
            if sep in key:
                prefix, _, algo = key.partition(sep)
                key = prefix + sep + algo.replace("'", "").replace("-", "").lower()
                break

        if key == "domain-name:value":
            # We force domain names to lower case.
            self._add_indicator(
                ioc=value.lower(),
                ioc_coll=collection,
                ioc_coll_list=collection["domains"],
            )
        elif key == "ipv4-addr:value":
            # We treat IP addresses as simple domains here to ease checks.
            self._add_indicator(
                ioc=value.strip(),
                ioc_coll=collection,
                ioc_coll_list=collection["domains"],
            )
        elif key == "process:name":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["processes"]
            )
        elif key == "email-addr:value":
            # We force email addresses to lower case.
            self._add_indicator(
                ioc=value.lower(),
                ioc_coll=collection,
                ioc_coll_list=collection["emails"],
            )
        elif key == "file:name":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["file_names"]
            )
        elif key == "file:path":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["file_paths"]
            )
        elif key == "file:hashes.md5":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["files_md5"]
            )
        elif key == "file:hashes.sha1":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["files_sha1"]
            )
        elif key == "file:hashes.sha256":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["files_sha256"]
            )
        elif key == "app:cert.md5":
            self._add_indicator(
                ioc=value,
                ioc_coll=collection,
                ioc_coll_list=collection["app_cert_hashes"],
            )
        elif key == "app:cert.sha1":
            self._add_indicator(
                ioc=value,
                ioc_coll=collection,
                ioc_coll_list=collection["app_cert_hashes"],
            )
        elif key == "app:cert.sha256":
            self._add_indicator(
                ioc=value,
                ioc_coll=collection,
                ioc_coll_list=collection["app_cert_hashes"],
            )
        elif key == "app:id":
            self._add_indicator(
                ioc=value, ioc_coll=collection, ioc_coll_list=collection["app_ids"]
            )
        elif key == "configuration-profile:id":
            self._add_indicator(
                ioc=value,
                ioc_coll=collection,
                ioc_coll_list=collection["ios_profile_ids"],
            )
        elif key == "android-property:name":
            self._add_indicator(
                ioc=value,
                ioc_coll=collection,
                ioc_coll_list=collection["android_property_names"],
            )
        elif key == "url:value":
            self._add_indicator(
                ioc=value,
                ioc_coll=collection,
                ioc_coll_list=collection["urls"],
            )
        else:
            self.log.debug("Can't add indicator %s, type %s not supported", value, key)

    def parse_stix2(self, file_path: str) -> None:
        """Extract indicators from a STIX2 file.

        :param file_path: Path to the STIX2 file to parse
        :type file_path: str

        """
        self.log.info("Parsing STIX2 indicators file at path '%s'", file_path)

        with open(file_path, "r", encoding="utf-8") as handle:
            try:
                data = json.load(handle)
            except json.decoder.JSONDecodeError:
                self.log.warning(
                    "Unable to parse STIX2 indicator file. "
                    "The file is corrupted or in the wrong format!"
                )
                return

        malware = {}
        indicators = []
        relationships = []
        reports = []
        for entry in data.get("objects", []):
            entry_type = entry.get("type", "")
            # Consider both malware and reports as collections
            if entry_type == "malware":
                malware[entry["id"]] = {
                    "name": entry["name"],
                    "description": entry.get("description", ""),
                }
            elif entry_type == "report":
                reports.append(entry)
            elif entry_type == "indicator":
                indicators.append(entry)
            elif entry_type == "relationship":
                relationships.append(entry)

        collections = []
        for mal_id, mal_values in malware.items():
            collection = self._new_collection(
                mal_id,
                mal_values.get("name"),
                mal_values.get("description"),
                os.path.basename(file_path),
                file_path,
            )
            collections.append(collection)

        for report in reports:
            collection = self._new_collection(
                report["id"],
                report.get("name", ""),
                report.get("description", ""),
                os.path.basename(file_path),
                file_path,
            )
            collections.append(collection)

        # Adds a default collection
        default_collection = self._new_collection(
            "0",
            "Default collection",
            "Collection with IOCs unrelated to malware or reports",
            os.path.basename(file_path),
            file_path,
        )

        # We loop through all indicators.
        for indicator in indicators:
            malware_id = None

            # We loop through reports first to see if the indicator is in the refs
            for report in reports:
                for ref in report.get("object_refs", []):
                    if ref == indicator["id"]:
                        malware_id = report["id"]
                        break

            if malware_id is None:
                # We loop through all relationships and find the one pertinent to
                # the current indicator.
                for relationship in relationships:
                    if relationship["source_ref"] != indicator["id"]:
                        continue

                    # Look for a malware definition with the correct identifier.
                    if relationship["target_ref"] in malware.keys():
                        malware_id = relationship["target_ref"]
                        break

            if malware_id is not None:
                # Now we look for the correct collection matching the malware ID we
                # got from the relationship.
                for collection in collections:
                    if collection["id"] == malware_id:
                        self._process_indicator(indicator, collection)
                        break
            else:
                # Adds to the default collection
                self._process_indicator(indicator, default_collection)

        for coll in collections:
            self.log.debug(
                'Extracted %d indicators for collection with name "%s"',
                coll["count"],
                coll["name"],
            )

        self.ioc_collections.extend(collections)
        if default_collection["count"] > 0:
            # Adds the default collection only if therare some IOCs in it
            self.ioc_collections.append(default_collection)

    def load_indicators_files(
        self, files: list, load_default: Optional[bool] = True
    ) -> None:
        """
        Load a list of indicators files.
        """
        for file_path in files:
            if os.path.isfile(file_path):
                self.parse_stix2(file_path)
            else:
                self.log.error("No indicators file exists at path %s", file_path)

        # Load downloaded indicators and any indicators from env variable.
        if load_default:
            self._load_downloaded_indicators()

        self._check_stix2_env_variable()
        self.log.info("Loaded a total of %d unique indicators", self.total_ioc_count)

    def get_iocs(self, ioc_type: str) -> Iterator[Indicator]:
        for ioc_collection in self.ioc_collections:
            for ioc in ioc_collection.get(ioc_type, []):
                yield Indicator(
                    value=ioc,
                    type=ioc_type,
                    name=ioc_collection["name"],
                    stix2_file_name=ioc_collection["stix2_file_name"],
                )

    @lru_cache()
    def get_ioc_matcher(
        self, ioc_type: Optional[str] = None, ioc_list: Optional[List[Indicator]] = None
    ) -> ahocorasick.Automaton:
        """
        Build an Aho-Corasick automaton from a list of iocs (i.e indicators)
        Returns an Aho-Corasick automaton

        This data-structue and algorithm allows for fast matching of a large number
        of match strings (i.e IOCs) against a large body of text. This will also
        match strings containing the IOC, so it is important to confirm the
        match is a valid IOC before using it.

            for _, ioc in domains_automaton.iter(url.domain.lower()):
                if ioc.value == url.domain.lower():
                    print(ioc)

        We use an LRU cache to avoid rebuilding the automaton every time we call a
        function such as check_url().
        """
        automaton = ahocorasick.Automaton()
        if ioc_type:
            iocs: Iterator[Indicator] = self.get_iocs(ioc_type)
        elif ioc_list:
            iocs = iter(ioc_list)
        else:
            raise ValueError("Must provide either ioc_type or ioc_list")

        for ioc in iocs:
            automaton.add_word(ioc.value, ioc)
        automaton.make_automaton()
        return automaton

    @lru_cache()
    def check_url(self, url: str) -> Optional[IndicatorMatch]:
        """Check if a given URL matches any of the provided domain indicators.

        :param url: URL to match against domain indicators
        :type url: str
        :returns: Indicator details if matched, otherwise None

        """
        if not url or not isinstance(url, str):
            return None

        # Check the URL first
        for ioc in self.get_iocs("urls"):
            if ioc.value == url:
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious URL {url} matching indicator "{ioc.value}" from "{ioc.name}"',
                )

        # Then check the domain
        # Create an Aho-Corasick automaton from the list of urls
        domain_matcher = self.get_ioc_matcher("domains")
        try:
            # First we use the provided URL.
            orig_url = URL(url)

            if orig_url.check_if_shortened():
                # If it is, we try to retrieve the actual URL making an
                # HTTP HEAD request.
                unshortened = orig_url.unshorten()

                self.log.debug("Found a shortened URL %s -> %s", url, unshortened)
                if unshortened is None:
                    self.log.warning("Unable to unshorten URL %s", url)
                    return None

                # Now we check for any nested URL shorteners.
                dest_url = URL(unshortened)
                if dest_url.check_if_shortened():
                    self.log.debug(
                        "Original URL %s appears to shorten another "
                        "shortened URL %s ... checking!",
                        orig_url.url,
                        dest_url.url,
                    )
                    match = self.check_url(dest_url.url)
                    self.resolved_urls[url] = self.resolved_urls.get(
                        dest_url.url, dest_url.url
                    )
                    return match

                final_url = dest_url
                self.resolved_urls[url] = final_url.url
            else:
                # If it's not shortened, we just use the original URL object.
                final_url = orig_url
        except Exception:
            # If URL parsing failed, we just try to do a simple substring
            # match.
            for _, ioc in domain_matcher.iter(url):
                if ioc.value.lower() in url:
                    return IndicatorMatch(
                        ioc=ioc,
                        message=f'Maybe found a known suspicious domain {url} matching indicator "{ioc.value}" from "{ioc.name}"',
                    )

            # If nothing matched, we can quit here.
            return None

        # If all parsing worked, we start walking through available domain
        # indicators.
        for _, ioc in domain_matcher.iter(final_url.domain.lower()):
            # First we check the full domain.
            if final_url.domain.lower() == ioc.value:
                if orig_url.is_shortened and orig_url.url != final_url.url:
                    message = f'Found a known suspicious domain {final_url.url} shortened as {orig_url.url} matching indicator "{ioc.value}" from "{ioc.name}"'
                else:
                    message = f'Found a known suspicious domain {final_url.url} matching indicator "{ioc.value}" from "{ioc.name}"'

                return IndicatorMatch(ioc=ioc, message=message)

        # Then we just check the top level domain.
        for _, ioc in domain_matcher.iter(final_url.top_level.lower()):
            if final_url.top_level.lower() == ioc.value:
                if orig_url.is_shortened and orig_url.url != final_url.url:
                    message = f'Found a sub-domain with suspicious top level {final_url.url} shortened as {orig_url.url} matching indicator "{ioc.value}" from "{ioc.name}"'
                else:
                    message = f'Found a sub-domain with a suspicious top level {final_url.url} matching indicator "{ioc.value}" from "{ioc.name}"'

                return IndicatorMatch(ioc=ioc, message=message)

        return None

    def get_expanded_url(self, url: str) -> Optional[str]:
        """Return the final URL recorded while checking a shortened URL."""
        expanded_url = self.resolved_urls.get(url)
        if expanded_url and expanded_url != url:
            return expanded_url
        return None

    def check_urls(self, urls: list) -> Optional[IndicatorMatch]:
        """Check a list of URLs against the provided list of domain indicators.

        :param urls: List of URLs to check against domain indicators
        :type urls: list
        :returns: Indicator details if matched, otherwise None

        """
        if not urls:
            return None

        return self.check_url_batches([urls])[0]

    def check_url_batches(
        self, url_batches: Sequence[Optional[Sequence[str]]]
    ) -> List[Optional[IndicatorMatch]]:
        """Check batches of URLs concurrently while preserving batch order.

        URLs are deduplicated across batches before checking. Each returned item
        is the first indicator match from the corresponding input batch, using
        the original URL order.
        """
        batches = [list(urls) if urls else [] for urls in url_batches]
        unique_urls = list(
            dict.fromkeys(url for urls in batches for url in urls)
        )

        if not unique_urls:
            return [None] * len(batches)

        if settings.NETWORK_ACCESS_ALLOWED and len(unique_urls) > 1:
            worker_count = min(URL_CHECK_MAX_WORKERS, len(unique_urls))
            with ThreadPoolExecutor(max_workers=worker_count) as executor:
                url_matches = dict(
                    zip(unique_urls, executor.map(self.check_url, unique_urls))
                )
        else:
            url_matches = {url: self.check_url(url) for url in unique_urls}

        return [
            next(
                (url_matches[url] for url in urls if url_matches[url] is not None),
                None,
            )
            for urls in batches
        ]

    def check_process(self, process: str) -> Optional[IndicatorMatch]:
        """Check the provided process name against the list of process
        indicators.

        :param process: Process name to check against process indicators
        :type process: str
        :returns: Indicator details if matched, otherwise None

        """
        if not process:
            return None

        proc_name = os.path.basename(process)
        for ioc in self.get_iocs("processes"):
            if proc_name == ioc.value:
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious process name "{process}" matching indicators from "{ioc.name}"',
                )

            if len(proc_name) == 16:
                if ioc.value.startswith(proc_name):
                    return IndicatorMatch(
                        ioc=ioc,
                        message=f'Found a truncated known suspicious process name "{process}" matching indicators from "{ioc.name}"',
                    )

        return None

    def check_processes(self, processes: list) -> Optional[IndicatorMatch]:
        """Check the provided list of processes against the list of
        process indicators.

        :param processes: List of processes to check against process indicators
        :type processes: list
        :returns: Indicator details if matched, otherwise None

        """
        if not processes:
            return None

        for process in processes:
            check = self.check_process(process)
            if check:
                return check

        return None

    def check_email(self, email: str) -> Optional[IndicatorMatch]:
        """Check the provided email against the list of email indicators.

        :param email: Email address to check against email indicators
        :type email: str
        :returns: Indicator details if matched, otherwise None

        """
        if not email:
            return None

        for ioc in self.get_iocs("emails"):
            if email.lower() == ioc.value.lower():
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious email address "{email}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_file_name(self, file_name: str) -> Optional[IndicatorMatch]:
        """Check the provided file name against the list of file indicators.

        :param file_name: File name to check against file
        indicators
        :type file_name: str
        :returns: Indicator details if matched, otherwise None

        """
        if not file_name:
            return None

        for ioc in self.get_iocs("file_names"):
            if ioc.value == file_name:
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious file name "{file_name}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_file_path(self, file_path: str) -> Optional[IndicatorMatch]:
        """Check the provided file path against the list of file indicators
        (both path and name).

        :param file_path: File path or file name to check against file
        indicators
        :type file_path: str
        :returns: Indicator details if matched, otherwise None

        """
        if not file_path:
            return None

        ioc_match = self.check_file_name(os.path.basename(file_path))
        if ioc_match:
            return ioc_match

        for ioc in self.get_iocs("file_paths"):
            # Strip any trailing slash from indicator paths to match
            # directories.
            if file_path.startswith(ioc.value.rstrip("/")):
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious file path "{file_path}" matching indicators form "{ioc.name}"',
                )

        return None

    def check_file_path_process(self, file_path: str) -> Optional[IndicatorMatch]:
        """Check the provided file path contains a process name from the
        list of indicators

        :param file_path: File path or file name to check against file
        indicators
        :type file_path: str
        :returns: Indicator details if matched, otherwise None

        """
        if not file_path:
            return None

        for ioc in self.get_iocs("processes"):
            parts = file_path.split("/")
            if ioc.value in parts:
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found known suspicious process name mentioned in file at path "{file_path}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_profile(self, profile_uuid: str) -> Optional[IndicatorMatch]:
        """Check the provided configuration profile UUID against the list of
        indicators.

        :param profile_uuid: Profile UUID to check against configuration profile
                             indicators
        :type profile_uuid: str
        :returns: Indicator details if matched, otherwise None

        """
        if not profile_uuid:
            return None

        for ioc in self.get_iocs("ios_profile_ids"):
            if profile_uuid in ioc.value:
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious profile ID "{profile_uuid}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_file_hash(self, file_hash: str) -> Optional[IndicatorMatch]:
        """Check the provided file hash against the list of indicators.

        :param file_hash: hash to check
        :type file_hash: str
        :returns: Indicator details if matched, otherwise None

        """
        if not file_hash:
            return None

        if len(file_hash) == 32:
            hash_type = "md5"
        elif len(file_hash) == 40:
            hash_type = "sha1"
        else:
            hash_type = "sha256"

        for ioc in self.get_iocs("files_" + hash_type):
            if file_hash.lower() == ioc.value.lower():
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious file with hash "{file_hash}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_app_certificate_hash(self, cert_hash: str) -> Optional[IndicatorMatch]:
        """Check the provided cert hash against the list of indicators.

        :param cert_hash: hash to check
        :type cert_hash: str
        :returns: Indicator details if matched, otherwise None

        """
        if not cert_hash:
            return None

        for ioc in self.get_iocs("app_cert_hashes"):
            if cert_hash.lower() == ioc.value.lower():
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious app certfificate with hash "{cert_hash}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_app_id(self, app_id: str) -> Optional[IndicatorMatch]:
        """Check the provided app identifier (typically an Android package name)
        against the list of indicators.

        :param app_id: App ID to check against the list of indicators
        :type app_id: str
        :returns: Indicator details if matched, otherwise None

        """
        if not app_id:
            return None

        for ioc in self.get_iocs("app_ids"):
            if app_id.lower() == ioc.value.lower():
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious app with ID "{app_id}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_android_property_name(
        self, property_name: str
    ) -> Optional[IndicatorMatch]:
        """Check the android property name against the list of indicators.

        :param property_name: Name of the Android property
        :type property_name: str
        :returns: Indicator details if matched, otherwise None

        """
        if property_name is None:
            return None

        for ioc in self.get_iocs("android_property_names"):
            if property_name.lower() == ioc.value.lower():
                return IndicatorMatch(
                    ioc=ioc,
                    message=f'Found a known suspicious Android property "{property_name}" matching indicators from "{ioc.name}"',
                )

        return None

    def check_domain(self, url: str) -> Optional[IndicatorMatch]:
        """
        Renamed check_url now, kept for compatibility
        """
        return self.check_url(url)

    def check_domains(self, urls: list) -> Optional[IndicatorMatch]:
        """
        Renamed check_domains, kept for compatibility
        """
        return self.check_urls(urls)



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import sqlite3
from base64 import b64encode
from typing import Optional

from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleResults,
    ModuleSerializedResult,
)
from mvt.common.utils import convert_mactime_to_iso

from ..base import IOSExtraction

SMS_BACKUP_IDS = [
    "3d0d7e5fb2ce288813306e4d4636395e047a3d28",
]
SMS_ROOT_PATHS = [
    "private/var/mobile/Library/SMS/sms.db",
]


class SMSAttachments(IOSExtraction):
    """This module extracts all info about SMS/iMessage attachments."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        return {
            "timestamp": record["isodate"],
            "module": self.__class__.__name__,
            "event": "sms_attachment",
            "data": f"{record['service']}: Attachment "
            f"'{record['transfer_name']}' {record['direction']} "
            f"from {record['phone_number']} "
            f"with {record['total_bytes']} bytes "
            f"(is_sticker: {record['is_sticker']}, "
            f"has_user_info: {record['has_user_info']})",
        }

    def check_indicators(self) -> None:
        for attachment in self.results:
            # Check for known malicious filenames.
            if self.indicators:
                ioc_match = self.indicators.check_file_path(attachment["filename"])
                if ioc_match:
                    self.alertstore.high(
                        ioc_match.message,
                        "",
                        attachment,
                        matched_indicator=ioc_match.ioc,
                    )

            if (
                attachment["filename"].startswith("/var/tmp/")
                and attachment["filename"].endswith("-1")
                and attachment["direction"] == "received"
            ):
                self.alertstore.medium(
                    f"Suspicious iMessage attachment {attachment['filename']} on {attachment['isodate']}",
                    attachment["isodate"],
                    attachment,
                )

    def run(self) -> None:
        self._find_ios_database(backup_ids=SMS_BACKUP_IDS, root_paths=SMS_ROOT_PATHS)
        self.log.info("Found SMS database at path: %s", self.file_path)

        if not self.file_path:
            return
        conn = self._open_sqlite_db(self.file_path)
        cur = conn.cursor()
        try:
            cur.execute(
                """
                SELECT
                    attachment.ROWID as "attachment_id",
                    attachment.*,
                    message.service as "service",
                    handle.id as "phone_number"
                FROM attachment
                LEFT JOIN message_attachment_join ON
                    message_attachment_join.attachment_id = attachment.ROWID
                LEFT JOIN message ON
                    message.ROWID = message_attachment_join.message_id
                LEFT JOIN handle ON handle.ROWID = message.handle_id;
            """
            )
        except sqlite3.OperationalError as exc:
            self.log.info(
                "No SMS attachment tables found in the database, skipping: %s", exc
            )
            cur.close()
            conn.close()
            return
        names = [description[0] for description in cur.description]

        for item in cur:
            attachment = {}
            for index, value in enumerate(item):
                if (
                    names[index]
                    in [
                        "user_info",
                        "sticker_user_info",
                        "attribution_info",
                        "ck_server_change_token_blob",
                        "sr_ck_server_change_token_blob",
                    ]
                ) and value:
                    value = b64encode(value).decode()
                attachment[names[index]] = value

            attachment["isodate"] = convert_mactime_to_iso(attachment["created_date"])
            attachment["start_date"] = convert_mactime_to_iso(attachment["start_date"])
            attachment["direction"] = (
                "sent" if attachment["is_outgoing"] == 1 else "received"
            )
            attachment["has_user_info"] = attachment["user_info"] is not None
            attachment["service"] = attachment["service"] or "Unknown"
            attachment["filename"] = attachment["filename"] or "NULL"
            self.results.append(attachment)

        cur.close()
        conn.close()

        self.log.info("Extracted a total of %d SMS attachments", len(self.results))



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import sqlite3
from typing import Optional

from mvt.common.module_types import ModuleResults

from ..base import IOSExtraction

CONTACTS_BACKUP_IDS = [
    "31bb7ba8914766d4ba40d6dfb6113c8b614be442",
]
CONTACTS_ROOT_PATHS = [
    "private/var/mobile/Library/AddressBook/AddressBook.sqlitedb",
]


class Contacts(IOSExtraction):
    """This module extracts all contact details from the phone's address book."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def run(self) -> None:
        self._find_ios_database(
            backup_ids=CONTACTS_BACKUP_IDS, root_paths=CONTACTS_ROOT_PATHS
        )
        self.log.info("Found Contacts database at path: %s", self.file_path)

        if not self.file_path:
            return
        conn = self._open_sqlite_db(self.file_path)
        cur = conn.cursor()
        try:
            try:
                cur.execute(
                    """
                    SELECT
                        multi.value, person.first, person.middle, person.last,
                        person.organization
                    FROM ABPerson person, ABMultiValue multi
                    WHERE person.rowid = multi.record_id and multi.value not null
                    ORDER by person.rowid ASC;
                """
                )
            except sqlite3.OperationalError as e:
                self.log.info("Error while reading the contact table: %s", e)
                return None
            names = [description[0] for description in cur.description]

            for row in cur:
                new_contact = {}
                for index, value in enumerate(row):
                    new_contact[names[index]] = value

                self.results.append(new_contact)
        finally:
            cur.close()
            conn.close()

        self.log.info(
            "Extracted a total of %d contacts from the address book", len(self.results)
        )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

from datetime import datetime
import re
from typing import Any

from mvt.common.module_types import ModuleAtomicResult, ModuleSerializedResult
from mvt.common.utils import convert_datetime_to_iso

from .artifact import AndroidArtifact

RISKY_PERMISSIONS = ["REQUEST_INSTALL_PACKAGES"]
RISKY_PACKAGES = ["com.android.shell"]


class DumpsysAppopsArtifact(AndroidArtifact):
    """
    Parser for dumpsys app ops info
    """

    def serialize(self, result: ModuleAtomicResult) -> ModuleSerializedResult:
        records = []
        for perm in result["permissions"]:
            if "entries" not in perm:
                continue

            for entry in perm["entries"]:
                if entry.get("timestamp"):
                    records.append(
                        {
                            "timestamp": entry["timestamp"],
                            "module": self.__class__.__name__,
                            "event": entry["event"],
                            "data": f"{result['package_name']} access to "
                            f"{perm['name']}: {entry['event']}",
                        }
                    )

        return records

    def check_indicators(self) -> None:
        for result in self.results:
            if self.indicators:
                ioc_match = self.indicators.check_app_id(result.get("package_name"))
                if ioc_match:
                    self.alertstore.critical(
                        ioc_match.message, "", result, matched_indicator=ioc_match.ioc
                    )
                    continue

            # We use a placeholder entry to create a basic alert even without permission entries.
            placeholder_entry = {"event": "unknown", "timestamp": ""}

            for perm in result["permissions"]:
                if (
                    perm["name"] in RISKY_PERMISSIONS
                    # and perm["access"] == "allow"
                ):
                    for entry in sorted(
                        perm["entries"] or [placeholder_entry],
                        key=lambda x: x.get("timestamp") or "",
                    ):
                        cleaned_result = result.copy()
                        cleaned_result["permissions"] = [perm]
                        self.alertstore.medium(
                            f"Package '{result['package_name']}' had risky permission '{perm['name']}' set to '{entry['event']}' at {entry['timestamp']}",
                            entry["timestamp"],
                            cleaned_result,
                        )

                elif result["package_name"] in RISKY_PACKAGES:
                    for entry in sorted(
                        perm["entries"] or [placeholder_entry],
                        key=lambda x: x.get("timestamp") or "",
                    ):
                        cleaned_result = result.copy()
                        cleaned_result["permissions"] = [perm]
                        self.alertstore.medium(
                            f"Risky package '{result['package_name']}' had '{perm['name']}' permission set to '{entry['event']}' at {entry['timestamp']}",
                            entry["timestamp"],
                            cleaned_result,
                        )

    def parse(self, output: str) -> None:
        self.results: list[dict[str, Any]] = []
        permission: dict[str, Any] | None = None
        package: dict[str, Any] | None = None
        uid: str | None = None
        uid_details: dict[str, Any] = {}
        attribution: str | None = None
        in_packages = False

        def finish_permission() -> None:
            nonlocal permission
            if package is not None and permission is not None:
                package["permissions"].append(permission)
            permission = None

        def finish_package() -> None:
            nonlocal package
            finish_permission()
            if package is not None:
                self.results.append(package)
            package = None

        for line in output.splitlines():
            uid_match = re.match(r"^  Uid ([^:]+):$", line)
            if uid_match:
                in_packages = True
                finish_package()
                uid = uid_match.group(1)
                uid_details = {
                    "uid_state": None,
                    "capability": None,
                    "app_widget_visible": None,
                    "default_modes": {},
                }
                continue

            if not in_packages:
                continue

            uid_property = re.match(
                r"^    (state|capability|appWidgetVisible)=(.*)$", line
            )
            if uid_property:
                key = {
                    "state": "uid_state",
                    "appWidgetVisible": "app_widget_visible",
                }.get(uid_property.group(1), uid_property.group(1))
                value: Any = uid_property.group(2)
                if value in ("true", "false"):
                    value = value == "true"
                uid_details[key] = value
                continue

            default_mode = re.match(r"^      ([A-Z0-9_]+): mode=([^\s]+)", line)
            if default_mode and package is None:
                uid_details["default_modes"][default_mode.group(1)] = (
                    default_mode.group(2)
                )
                continue

            if line.startswith("    Package "):
                finish_package()
                package = {
                    "package_name": line[12:-1],
                    "permissions": [],
                    "uid": uid,
                    **uid_details,
                }
                continue

            operation_match = re.match(
                r"^      ([A-Z0-9_]+)(?: \(([^)]+)\))?:\s*$", line
            )
            if package is not None and operation_match:
                finish_permission()
                permission = {
                    "name": operation_match.group(1),
                    "mode": operation_match.group(2),
                    "entries": [],
                }
                attribution = None
                continue

            attribution_match = re.match(r"^\s{8,}([^=]+)=\[$", line)
            if attribution_match:
                attribution = attribution_match.group(1).strip()
                continue
            if line.strip() == "]":
                attribution = None
                continue

            if permission is None:
                continue
            event_match = re.match(
                r"^\s*(Access|Reject):\s*\[([^]]+)\]\s*"
                r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+)\s*"
                r"(\([^)]*\))?(?:\s+duration=([^\s]+))?",
                line,
            )
            running_match = re.match(
                r"^\s*Running start at:\s*(\S+(?: \S+)?)",
                line,
            )
            if event_match:
                entry = {
                    "event": event_match.group(1).lower(),
                    "access": event_match.group(1),
                    "uid_state": event_match.group(2),
                    "timestamp": convert_datetime_to_iso(
                        datetime.strptime(event_match.group(3), "%Y-%m-%d %H:%M:%S.%f")
                    ),
                    "relative_time": event_match.group(4),
                    "duration": event_match.group(5),
                    "attribution": attribution,
                }
                permission["entries"].append(entry)
            elif running_match:
                raw_start = running_match.group(1)
                timestamp = None
                if re.fullmatch(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+", raw_start):
                    timestamp = convert_datetime_to_iso(
                        datetime.strptime(raw_start, "%Y-%m-%d %H:%M:%S.%f")
                    )
                permission["entries"].append(
                    {
                        "event": "running",
                        "access": "Running",
                        "uid_state": None,
                        "timestamp": timestamp,
                        "relative_time": raw_start
                        if raw_start.startswith("+")
                        else None,
                        "duration": None,
                        "attribution": attribution,
                    }
                )

        finish_package()



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

# Help messages of generic options.
HELP_MSG_VERSION = "Show the currently installed version of MVT"
HELP_MSG_OUTPUT = "Specify a path to a folder where you want to store JSON results"
HELP_MSG_IOC = "Path to indicators file (can be invoked multiple time)"
HELP_MSG_FAST = "Avoid running time/resource consuming features"
HELP_MSG_LIST_MODULES = "Print list of available modules and their source, then exit"
HELP_MSG_MODULE = "Name of a single module you would like to run instead of all"
HELP_MSG_LOAD_MODULE = (
    "Load custom MVT module(s) from a Python file or folder "
    "(can be invoked multiple times)"
)
HELP_MSG_NONINTERACTIVE = "Don't ask interactive questions during processing"
HELP_MSG_HASHES = "Generate hashes of all the files analyzed"
HELP_MSG_VERBOSE = "Verbose mode"
HELP_MSG_VERBOSE_COMMAND = (
    "Verbose mode (kept for compatibility, pass --verbose before the command "
    "name instead)"
)
HELP_MSG_CHECK_IOCS = "Compare stored JSON results to provided indicators"
HELP_MSG_STIX2 = "Download public STIX2 indicators"
HELP_MSG_DISABLE_UPDATE_CHECK = "Disable MVT version update check"
HELP_MSG_DISABLE_INDICATOR_UPDATE_CHECK = "Disable indicators update check"
HELP_MSG_COMPLETION = (
    "Generate or install shell completion for mvt, mvt-ios and mvt-android"
)
HELP_MSG_PLUGINS = "Inspect the installed MVT plugin packages"
HELP_MSG_PLUGINS_LIST = "List the installed plugins and what they contribute to MVT"
HELP_MSG_PLUGINS_CHECK_UPDATES = (
    "Check the installed plugins for updates without installing them"
)

# IOS Specific
HELP_MSG_DECRYPT_BACKUP = "Decrypt an encrypted iTunes backup"
HELP_MSG_BACKUP_DESTINATION = (
    "Path to the folder where the decrypted backup should be stored"
)
HELP_MSG_IOS_BACKUP_PASSWORD = (
    "Password to use to decrypt the backup (or, set the {MVT_IOS_BACKUP_PASSWORD} "
    "environment variable)"
)
HELP_MSG_BACKUP_KEYFILE = (
    "File containing raw encryption key to use to decrypt the backup"
)
HELP_MSG_EXTRACT_KEY = "Extract decryption key from an iTunes backup"
HELP_MSG_CHECK_IOS_BACKUP = "Extract artifacts from an iTunes backup"
HELP_MSG_CHECK_FS = "Extract artifacts from a full filesystem dump"
HELP_MSG_CHECK_SYSDIAGNOSE = "Extract artifacts from an iOS sysdiagnose archive"

# Android Specific
HELP_MSG_ANDROID_BACKUP_PASSWORD = "The backup password to use for an Android backup"
HELP_MSG_CHECK_ADB_REMOVED = "REMOVED: Check an Android device over ADB"
HELP_MSG_CHECK_ADB_REMOVED_DESCRIPTION = (
    "The 'mvt-android check-adb' command has been removed from MVT. "
    "Use AndroidQF to collect full forensic artifacts from an Android device. \n\n"
    "The 'mvt-android check-androidqf' command in MVT can be used to fully analyze "
    "forensic data collected with AndroidQF. Minimal checks can also be performed "
    "on an Android bugreport using the 'mvt-android check-bugreport' command."
)
HELP_MSG_CHECK_BUGREPORT = "Check an Android Bug Report"
HELP_MSG_CHECK_ANDROID_BACKUP = "Check an Android Backup"
HELP_MSG_CHECK_ANDROIDQF = "Check data collected with AndroidQF"
HELP_MSG_CHECK_INTRUSION_LOGS = "Check Android Intrusion Logging files"
HELP_MSG_VIRUS_TOTAL = "Check package APK hashes on VirusTotal"
HELP_MSG_DELAY_CHECKS = "Delay in seconds between VirusTotal requests"



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import logging
import operator
import sqlite3
from pathlib import Path
from typing import Optional

from mvt.common.utils import convert_mactime_to_iso

from .base import IOSExtraction
from mvt.common.module_types import (
    ModuleAtomicResult,
    ModuleSerializedResult,
    ModuleResults,
)


class NetBase(IOSExtraction):
    """This class provides a base for DataUsage and NetUsage extraction
    modules."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )

    def _extract_net_data(self):
        assert self.file_path is not None
        conn = self._open_sqlite_db(self.file_path)
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        try:
            cur.execute(
                """
                SELECT
                    ZPROCESS.ZFIRSTTIMESTAMP,
                    ZPROCESS.ZTIMESTAMP,
                    ZPROCESS.ZPROCNAME,
                    ZPROCESS.ZBUNDLENAME,
                    ZPROCESS.Z_PK AS ZPROCESS_PK,
                    ZLIVEUSAGE.ZWIFIIN,
                    ZLIVEUSAGE.ZWIFIOUT,
                    ZLIVEUSAGE.ZWWANIN,
                    ZLIVEUSAGE.ZWWANOUT,
                    ZLIVEUSAGE.Z_PK AS ZLIVEUSAGE_PK,
                    ZLIVEUSAGE.ZHASPROCESS,
                    ZLIVEUSAGE.ZTIMESTAMP AS ZL_TIMESTAMP
                FROM ZLIVEUSAGE
                LEFT JOIN ZPROCESS ON ZLIVEUSAGE.ZHASPROCESS = ZPROCESS.Z_PK
                UNION
                SELECT ZFIRSTTIMESTAMP, ZTIMESTAMP, ZPROCNAME, ZBUNDLENAME, Z_PK,
                    NULL, NULL, NULL, NULL, NULL, NULL, NULL
                FROM ZPROCESS WHERE Z_PK NOT IN
                    (SELECT ZHASPROCESS FROM ZLIVEUSAGE);
            """
            )
        except sqlite3.OperationalError:
            # Recent phones don't have ZWIFIIN and ZWIFIOUT columns
            cur.execute(
                """
                SELECT
                    ZPROCESS.ZFIRSTTIMESTAMP,
                    ZPROCESS.ZTIMESTAMP,
                    ZPROCESS.ZPROCNAME,
                    ZPROCESS.ZBUNDLENAME,
                    ZPROCESS.Z_PK AS ZPROCESS_PK,
                    ZLIVEUSAGE.ZWWANIN,
                    ZLIVEUSAGE.ZWWANOUT,
                    ZLIVEUSAGE.Z_PK AS ZLIVEUSAGE_PK,
                    ZLIVEUSAGE.ZHASPROCESS,
                    ZLIVEUSAGE.ZTIMESTAMP AS ZL_TIMESTAMP
                FROM ZLIVEUSAGE
                LEFT JOIN ZPROCESS ON ZLIVEUSAGE.ZHASPROCESS = ZPROCESS.Z_PK
                UNION
                SELECT ZFIRSTTIMESTAMP, ZTIMESTAMP, ZPROCNAME, ZBUNDLENAME, Z_PK,
                    NULL, NULL, NULL, NULL, NULL
                FROM ZPROCESS WHERE Z_PK NOT IN
                    (SELECT ZHASPROCESS FROM ZLIVEUSAGE);
            """
            )

        for row in cur:
            # ZPROCESS records can be missing after the JOIN.
            # Handle NULL timestamps.
            if row["ZFIRSTTIMESTAMP"] and row["ZTIMESTAMP"]:
                first_isodate = convert_mactime_to_iso(row["ZFIRSTTIMESTAMP"])
                isodate = convert_mactime_to_iso(row["ZTIMESTAMP"])
            else:
                first_isodate = row["ZFIRSTTIMESTAMP"]
                isodate = row["ZTIMESTAMP"]

            if row["ZL_TIMESTAMP"]:
                live_timestamp = convert_mactime_to_iso(row["ZL_TIMESTAMP"])
            else:
                live_timestamp = ""

            self.results.append(
                {
                    "first_isodate": first_isodate,
                    "isodate": isodate,
                    "proc_name": row["ZPROCNAME"],
                    "bundle_id": row["ZBUNDLENAME"],
                    "proc_id": row["ZPROCESS_PK"],
                    "wifi_in": row["ZWIFIIN"] if "ZWIFIIN" in row.keys() else None,
                    "wifi_out": row["ZWIFIOUT"] if "ZWIFIOUT" in row.keys() else None,
                    "wwan_in": row["ZWWANIN"],
                    "wwan_out": row["ZWWANOUT"],
                    "live_id": row["ZLIVEUSAGE_PK"],
                    "live_proc_id": row["ZHASPROCESS"],
                    "live_isodate": live_timestamp
                    if row["ZL_TIMESTAMP"]
                    else first_isodate,
                }
            )

        cur.close()
        conn.close()

        self.log.info("Extracted information on %d processes", len(self.results))

    def serialize(self, record: ModuleAtomicResult) -> ModuleSerializedResult:
        record_data = (
            f"{record['proc_name']} (Bundle ID: {record['bundle_id']},"
            f" ID: {record['proc_id']})"
        )
        record_data_usage = (
            record_data + " "
            f"WWAN IN: {record['wwan_in']}, "
            f"WWAN OUT: {record['wwan_out']}"
        )

        records = [
            {
                "timestamp": record["live_isodate"],
                "module": self.__class__.__name__,
                "event": "live_usage",
                "data": record_data_usage,
            }
        ]

        # Only included first_usage and current_usage records when a
        # ZPROCESS entry exists.
        if (
            "MANIPULATED" not in record["proc_name"]
            and "MISSING" not in record["proc_name"]
            and record["live_proc_id"] is not None
        ):
            records.extend(
                [
                    {
                        "timestamp": record["first_isodate"],
                        "module": self.__class__.__name__,
                        "event": "first_usage",
                        "data": record_data,
                    },
                    {
                        "timestamp": record["isodate"],
                        "module": self.__class__.__name__,
                        "event": "current_usage",
                        "data": record_data,
                    },
                ]
            )

        return records

    def _find_suspicious_processes(self):
        if not self.is_fs_dump:
            return

        if not self.results:
            return

        # If we are instructed to run fast, we skip this.
        if self.module_options.get("fast_mode", None):
            self.log.info(
                "Option 'fast_mode' was enabled: skipping extended "
                "search for suspicious processes"
            )
            return

        self.log.info("Extended search for suspicious processes ...")

        files = []
        for posix_path in Path(self.target_path).rglob("*"):
            try:
                if not posix_path.is_file():
                    continue
            except PermissionError:
                continue

            files.append([posix_path.name, str(posix_path)])

        for proc in self.results:
            if not proc["bundle_id"]:
                self.log.debug(
                    "Found process with no Bundle ID with name: %s", proc["proc_name"]
                )

                binary_path = None
                for file in files:
                    if proc["proc_name"] == file[0]:
                        binary_path = file[1]
                        break

                if binary_path:
                    self.log.debug("Located at %s", binary_path)
                else:
                    msg = (
                        "Could not find the binary associated with the "
                        f"process with name {proc['proc_name']}"
                    )
                    if not proc["proc_name"]:
                        msg = (
                            "Found process entry with empty 'proc_name': "
                            f"{proc['live_proc_id']} at {proc['live_isodate']}"
                        )
                    elif len(proc["proc_name"]) == 16:
                        msg += (
                            " (However, the process name might have "
                            "been truncated in the database)"
                        )

                    self.alertstore.medium(msg, proc["live_isodate"], proc)

            if not proc["live_proc_id"]:
                self.log.info(
                    "Found process entry in ZPROCESS but not in ZLIVEUSAGE: %s at %s",
                    proc["proc_name"],
                    proc["live_isodate"],
                )

    def check_manipulated(self):
        """Check for missing or manipulate DB entries"""
        # Don't show duplicates for each missing process.
        missing_process_cache = set()
        for result in sorted(self.results, key=operator.itemgetter("live_isodate")):
            if result["proc_id"]:
                continue

            # Avoid duplicate warnings for same process.
            if result["live_proc_id"] not in missing_process_cache:
                missing_process_cache.add(result["live_proc_id"])
                self.alertstore.high(
                    f"Found manipulated process entry {result['live_proc_id']}. Entry on {result['live_isodate']}",
                    result["live_isodate"],
                    result,
                )

            # Set manipulated proc timestamp so it appears in timeline.
            result["first_isodate"] = result["isodate"] = result["live_isodate"]
            result["proc_name"] = "MANIPULATED [process record deleted]"
            self.alertstore.high(
                f"Found manipulated process entry {result['live_proc_id']}/",
                result["first_isodate"],
                result,
            )

    def find_deleted(self):
        """Identify process which may have been deleted from the DataUsage
        database."""
        results_by_proc = {
            proc["proc_id"]: proc for proc in self.results if proc["proc_id"]
        }
        all_proc_id = sorted(results_by_proc.keys())

        # Fix issue #108
        if not all_proc_id:
            return

        missing_procs, last_proc_id = {}, None
        for proc_id in range(min(all_proc_id), max(all_proc_id)):
            if proc_id not in all_proc_id:
                previous_proc = results_by_proc[last_proc_id]
                self.alertstore.low(
                    f'Missing process {proc_id}. Previous process at "{previous_proc["first_isodate"]}" ({previous_proc["proc_name"]})',
                    previous_proc["first_isodate"],
                    previous_proc,
                )

                missing_procs[proc_id] = {
                    "proc_id": proc_id,
                    "prev_proc_id": last_proc_id,
                    "prev_proc_name": previous_proc["proc_name"],
                    "prev_proc_bundle": previous_proc["bundle_id"],
                    "prev_proc_first": previous_proc["first_isodate"],
                }
            else:
                last_proc_id = proc_id

        # Add a placeholder entry for the missing processes.
        for proc_id, proc in missing_procs.items():
            # Set default DataUsage keys.
            result = {key: None for key in self.results[0].keys()}
            result["first_isodate"] = result["isodate"] = result["live_isodate"] = proc[
                "prev_proc_first"
            ]
            result["proc_name"] = f"MISSING [follows {proc['prev_proc_name']}]"
            result["proc_id"] = result["live_proc_id"] = proc["proc_id"]
            result["bundle_id"] = None

            self.results.append(result)

        self.results = sorted(self.results, key=operator.itemgetter("first_isodate"))

    def check_indicators(self) -> None:
        # check_manipulated/find_deleted require "live_isodate" and
        # "live_proc_id" keys which may be absent in older result formats.
        if self.results and "live_isodate" in self.results[0]:
            self.check_manipulated()
            self.find_deleted()

        if not self.indicators:
            return

        for result in self.results:
            try:
                proc_name = result["proc_name"]
            except KeyError:
                continue

            # Process ID may be empty if process records have been manipulated.
            if not result["proc_id"]:
                continue

            ioc_match = self.indicators.check_process(proc_name)
            if ioc_match:
                self.alertstore.critical(
                    ioc_match.message,
                    result["first_isodate"],
                    result,
                    matched_indicator=ioc_match.ioc,
                )



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import base64
import binascii
import hashlib
import re

from .artifact import AndroidArtifact


class DumpsysADBArtifact(AndroidArtifact):
    multiline_fields = ["user_keys", "keystore"]

    @staticmethod
    def is_structural_line(key: str, vals: list) -> bool:
        if key == "}":
            return True
        # XML keystore continuations also split on "=", but never into an identifier.
        return len(vals) == 2 and key.isidentifier()

    def indented_dump_parser(self, dump_data):
        """
        Parse the indented dumpsys output, generated by DualDumpOutputStream in Android.
        """
        res = {}
        stack = [res]
        cur_indent = 0
        in_multiline = False
        for line in dump_data.strip(b"\n").split(b"\n"):
            # Track the level of indentation
            indent = len(line) - len(line.lstrip())
            if indent < cur_indent:
                # If the current line is less indented than the previous one, back out
                stack.pop()
                cur_indent = indent
            else:
                cur_indent = indent

            # Split key and value by '='
            vals = line.lstrip().split(b"=", 1)
            key = vals[0].decode("utf-8")
            current_dict = stack[-1]

            # Annoyingly, some values are multiline and don't have a key on each line
            if in_multiline:
                if key == "":
                    # If the line is empty, it's the terminator for the multiline value
                    in_multiline = False
                    if isinstance(stack[-1], list):
                        stack.pop()
                    continue

                if not self.is_structural_line(key, vals):
                    current_dict.append(line.lstrip())
                    continue

                in_multiline = False
                if isinstance(stack[-1], list):
                    stack.pop()
                current_dict = stack[-1]

            if key == "}":
                stack.pop()
                continue

            if vals[1] == b"{":
                # If the value is a new dictionary, add it to the stack
                current_dict[key] = {}
                stack.append(current_dict[key])

            # Handle continue multiline values
            elif key in self.multiline_fields:
                current_dict[key] = []
                current_dict[key].append(vals[1])

                in_multiline = True
                stack.append(current_dict[key])
            else:
                # If the value something else, store it in the current dictionary
                current_dict[key] = vals[1]

        return res

    def parse_xml(self, xml_data):
        """
        Parse XML data from dumpsys ADB output
        """
        import xml.etree.ElementTree as ET

        keystore = []
        keystore_root = ET.fromstring(xml_data)
        for adb_key in keystore_root.findall("adbKey"):
            key_info = self.calculate_key_info(adb_key.get("key").encode("utf-8"))
            key_info["last_connected"] = adb_key.get("lastConnection")
            keystore.append(key_info)

        return keystore

    def parse_binary_xml(self, data: bytes) -> list[dict]:
        """Recover ADB key records from Android binary XML (ABX).

        Some dumpstate implementations embed ABX in a text stream and replace
        binary token bytes. The public key remains intact, while unavailable
        numeric metadata is represented as ``None`` rather than corrupt text.
        """
        keystore = []
        seen = set()
        for match in re.finditer(
            rb"(?<![A-Za-z0-9+/])([A-Za-z0-9+/]{300,}={0,2})"
            rb"(?: ([A-Za-z0-9_.@-]+))?",
            data,
        ):
            key = match.group(1)
            try:
                base64.b64decode(key, validate=True)
            except (binascii.Error, ValueError):
                continue
            if key in seen:
                continue
            seen.add(key)
            full_key = key + (b" " + match.group(2) if match.group(2) else b"")
            key_info = self.calculate_key_info(full_key)
            key_info["last_connected"] = None
            keystore.append(key_info)
        return keystore

    @staticmethod
    def calculate_key_info(user_key: bytes) -> dict:
        if b" " in user_key:
            key_base64, user = user_key.split(b" ", 1)
        else:
            key_base64, user = user_key, b""

        try:
            key_raw = base64.b64decode(key_base64)
            key_fingerprint = hashlib.md5(key_raw).hexdigest().upper()
            key_fingerprint_colon = ":".join(
                [key_fingerprint[i : i + 2] for i in range(0, len(key_fingerprint), 2)]
            )
        except binascii.Error:
            # Impossible to parse base64
            key_fingerprint_colon = ""

        return {
            "user": user.decode("utf-8"),
            "fingerprint": key_fingerprint_colon,
            "key": key_base64.decode("ascii", errors="replace"),
        }

    def check_indicators(self) -> None:
        if not self.results:
            return

        for entry in self.results:
            for user_key in entry.get("user_keys", []):
                self.log.debug(
                    f"Found trusted ADB key for user '{user_key['user']}' with fingerprint "
                    f"'{user_key['fingerprint']}'"
                )

    def parse(self, content: bytes) -> None:
        """
        Parse the Dumpsys ADB section
        Adds results to self.results (List[Dict[str, str]])

        :param content: content of the ADB section (string)
        """
        if not content or b"Can't find service: adb" in content:
            self.log.error(
                "Could not load ADB data from dumpsys. "
                "It may not be supported on this device."
            )
            return

        start_of_json = content.find(b"\n{")
        if start_of_json == -1:
            self.log.error("Unable to find ADB manager state in dumpsys output")
            return

        end_of_json = content.rfind(b"}\n")
        if end_of_json == -1 or end_of_json <= start_of_json:
            self.log.error("Unable to find complete ADB manager state in dumpsys output")
            return

        json_content = content[start_of_json + 2 : end_of_json - 2].rstrip()

        parsed = self.indented_dump_parser(json_content)
        if parsed.get("debugging_manager") is None:
            self.log.error("Unable to find expected ADB entries in dumpsys output")  # noqa
            return

        # Keystore can be in different levels, as the basic parser
        # is not always consistent due to different dumpsys formats.
        if parsed.get("keystore"):
            keystore_data = b"\n".join(parsed["keystore"])
        elif parsed["debugging_manager"].get("keystore"):
            keystore_data = b"\n".join(parsed["debugging_manager"]["keystore"])
        else:
            keystore_data = None

        # Keystore is in XML format on some devices and we need to parse it
        if keystore_data and keystore_data.startswith(b"<?xml"):
            parsed["debugging_manager"]["keystore"] = self.parse_xml(keystore_data)
        elif keystore_data and keystore_data.startswith(b"ABX\x00"):
            parsed["debugging_manager"]["keystore"] = self.parse_binary_xml(
                keystore_data
            )
        else:
            # Keystore is not XML format
            parsed["debugging_manager"]["keystore"] = keystore_data

        parsed = parsed["debugging_manager"]

        for key, value in list(parsed.items()):
            if isinstance(value, bytes):
                decoded = value.decode("utf-8", errors="replace")
                parsed[key] = (
                    decoded == "true" if decoded in ("true", "false") else decoded
                )

        # Calculate key fingerprints for better readability
        key_info = []
        for user_key in parsed.get("user_keys", []):
            user_info = self.calculate_key_info(user_key)
            key_info.append(user_info)

        parsed["user_keys"] = key_info
        self.results = [parsed]



# Mobile Verification Toolkit (MVT)
# Copyright (c) 2021-2023 The MVT Authors.
# Use of this software is governed by the MVT License 1.1 that can be found at
#   https://license.mvt.re/1.1/

import fnmatch
import logging
import os
import re
import tarfile
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from mvt.common.module import MVTModule, ModuleResults


class SysdiagnoseExtraction(MVTModule):
    """Base class for custom modules that analyze an iOS sysdiagnose."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        target_path: Optional[str] = None,
        results_path: Optional[str] = None,
        module_options: Optional[dict] = None,
        log: logging.Logger = logging.getLogger(__name__),
        results: Optional[ModuleResults] = None,
    ) -> None:
        super().__init__(
            file_path=file_path,
            target_path=target_path,
            results_path=results_path,
            module_options=module_options,
            log=log,
            results=results,
        )
        self.parent_path: Optional[str] = None
        self.files: list[str] = []
        self.tar: Optional[tarfile.TarFile] = None
        self.tar_files: list[str] = []
        self.ips_files: list[dict[str, object]] = []

    def from_sysdiagnose_folder(
        self, target_path: str, sysdiagnose_files: list[str]
    ) -> None:
        self.parent_path = Path(target_path).absolute().parent.as_posix()
        self.files = sysdiagnose_files

    def from_sysdiagnose_tar(
        self, sysdiagnose_archive: tarfile.TarFile, sysdiagnose_files: list[str]
    ) -> None:
        self.tar = sysdiagnose_archive
        self.tar_files = sysdiagnose_files

    def _extract_timezone(self):
        """Determine the sysdiagnose timezone from its diagnostic log."""
        file_paths = self._get_files_by_pattern("*/sysdiagnose.log")
        if not file_paths:
            self.log.info(
                "Unable to determine the timezone in which the sysdiagnose was "
                "generated. Assuming UTC for logs without a timezone."
            )
            return timezone.utc

        content = self._get_file_content(file_paths[0]).decode(
            "utf-8", errors="replace"
        )
        filenames = re.findall(r"sysdiagnose_\S+?\.tar\.gz", content)
        if not filenames:
            self.log.info(
                "Unable to determine the timezone in which the sysdiagnose was "
                "generated. Assuming UTC for logs without a timezone."
            )
            return timezone.utc

        timestamp = "_".join(
            filenames[0].removesuffix(".tar.gz").split("_")[1:3]
        )
        sysdiagnose_timezone = datetime.strptime(
            timestamp, "%Y.%m.%d_%H-%M-%S%z"
        ).tzinfo
        self.log.info(
            "Based on the sysdiagnose filename, assuming timezone %s for logs "
            "without a timezone.",
            sysdiagnose_timezone,
        )
        return sysdiagnose_timezone

    def _get_files_by_pattern(self, pattern: str) -> list[str]:
        file_names = self.tar_files if self.tar else self.files
        return fnmatch.filter(file_names, pattern)

    def _get_file_content(self, file_path: str) -> bytes:
        if self.tar:
            handle = self.tar.extractfile(self.tar.getmember(file_path))
        else:
            if self.parent_path is None:
                raise RuntimeError("Sysdiagnose folder has not been initialized")
            handle = open(os.path.join(self.parent_path, file_path), "rb")

        if handle is None:
            return b""
        with handle:
            return handle.read()
